import unittest

import numpy as np

from kafe2.core.error import MatrixGaussianError, SimpleGaussianError


class TestMatrixGaussianError(unittest.TestCase):

    def setUp(self):
        self.ref_cov_mat = np.array([[0.10, 0.04, 0.29],
                                     [0.04, 0.02, 0.42],
                                     [0.29, 0.42, 0.33]])
        self.ref_cor_mat = np.array([[ 1.        ,  0.89442719,  1.59639746],
                                     [ 0.89442719,  1.        ,  5.16984262],
                                     [ 1.59639746,  5.16984262,  1.        ]])
        self.ref_err_val = np.sqrt(np.array([0.1, 0.02, 0.33]))
        self.ref_cov_mat_inverse = np.linalg.inv(self.ref_cov_mat)

        # -- errors with no reference

        # construct error from cov mat with no reference (absolute error)
        self.ge_cov_noref = MatrixGaussianError(self.ref_cov_mat, 'cov')
        # construct error from cor mat and error array, with no reference (absolute error)
        self.ge_cor_noref = MatrixGaussianError(self.ref_cor_mat, 'cor', err_val=self.ref_err_val)

        # -- errors with reference
        self.ref_reference = [0.1, 0.2, 0.3]
        self.ref_cov_mat_rel = np.array([[10.0,   2.0, 9.667],
                                         [ 2.0,   0.5, 7.0  ],
                                         [ 9.667, 7.0, 3.667]])
        self.ref_err_val_rel = np.array([3.16227766, 0.70710678, 1.91494125])

        # construct error from cov mat with reference (absolute error)
        self.ge_cov_wref = MatrixGaussianError(self.ref_cov_mat, 'cov', reference=self.ref_reference)
        # construct error from cor mat and error array, with reference (absolute error)
        self.ge_cor_wref = MatrixGaussianError(self.ref_cor_mat, 'cor', err_val=self.ref_err_val, reference=self.ref_reference)

        # -- relative errors

        # construct error from cov mat with no reference (relative error)
        self.ge_cov_rel_noref = MatrixGaussianError(self.ref_cov_mat_rel, 'cov', relative=True)
        # construct error from cor mat and error array, with no reference (relative error)
        self.ge_cor_rel_noref = MatrixGaussianError(self.ref_cor_mat, 'cor', err_val=self.ref_err_val_rel, relative=True)
        # construct error from cov mat with reference (relative error)
        self.ge_cov_rel_wref = MatrixGaussianError(self.ref_cov_mat_rel, 'cov', relative=True, reference=self.ref_reference)
        # construct error from cor mat and error array, with reference (relative error)
        self.ge_cor_rel_wref = MatrixGaussianError(self.ref_cor_mat, 'cor', err_val=self.ref_err_val_rel, relative=True, reference=self.ref_reference)

        # -- ill defined matrices
        self.wrong_cor_mat_diagonal_not_ones = np.array([[42., 0.89442719, 1.59639746],
                                                         [0.89442719, 1., 5.16984262],
                                                         [1.59639746, 5.16984262, 1.]])

        self.wrong_cor_mat_asymm = np.array([[1., 42.89442719, 1.59639746],
                                             [0.89442719, 1., 5.16984262],
                                             [1.59639746, 5.16984262, 1.]])

        self.wrong_cov_mat_asymm = np.array([[0.10, 42.0, 0.29],
                                             [0.04, 0.02, 0.42],
                                             [0.29, 0.42, 0.33]])

    def test_compare_cov_from_cov_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_wref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_compare_cov_rel_from_cov_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_rel_wref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_compare_cov_from_cor_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_wref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_compare_cov_rel_from_cor_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_rel_wref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_compare_error_from_cov_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_wref.error, self.ref_err_val, atol=1e-3))

    def test_compare_error_rel_from_cov_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_rel_wref.error_rel, self.ref_err_val_rel, atol=1e-3))

    def test_compare_error_from_cor_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_wref.error, self.ref_err_val, atol=1e-3))

    def test_compare_error_rel_from_cor_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_rel_wref.error_rel, self.ref_err_val_rel, atol=1e-3))

    def test_compare_cov_from_cov_noref(self):
        self.assertTrue(np.allclose(self.ge_cov_noref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_compare_cov_rel_from_cov_rel_noref(self):
        self.assertTrue(np.allclose(self.ge_cov_rel_noref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_compare_cov_from_cor_noref(self):
        self.assertTrue(np.allclose(self.ge_cor_noref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_compare_cov_rel_from_cor_rel_noref(self):
        self.assertTrue(np.allclose(self.ge_cor_rel_noref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_compare_error_from_cov_noref(self):
        self.assertTrue(np.allclose(self.ge_cov_noref.error, self.ref_err_val, atol=1e-3))

    def test_compare_error_rel_from_cov_rel_noref(self):
        self.assertTrue(np.allclose(self.ge_cov_rel_noref.error_rel, self.ref_err_val_rel, atol=1e-3))

    def test_compare_error_from_cor_noref(self):
        self.assertTrue(np.allclose(self.ge_cor_noref.error, self.ref_err_val, atol=1e-3))

    def test_compare_error_rel_from_cor_rel_noref(self):
        self.assertTrue(np.allclose(self.ge_cor_rel_noref.error_rel, self.ref_err_val_rel, atol=1e-3))

    def test_compare_cov_mat_inverse(self):
        self.assertTrue(np.allclose(self.ge_cov_noref.cov_mat_inverse, self.ref_cov_mat_inverse))
        self.assertTrue(np.allclose(self.ge_cov_wref.cov_mat_inverse, self.ref_cov_mat_inverse))
        self.assertTrue(np.allclose(self.ge_cor_noref.cov_mat_inverse, self.ref_cov_mat_inverse))
        self.assertTrue(np.allclose(self.ge_cor_wref.cov_mat_inverse, self.ref_cov_mat_inverse))

    def test_convert_cov_from_cov_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_rel_wref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_convert_cov_rel_from_cov_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_wref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_convert_cov_from_cor_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_rel_wref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_convert_cov_rel_from_cor_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_wref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_convert_error_from_cov_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_rel_wref.error, self.ref_err_val, atol=1e-3))

    def test_convert_error_rel_from_cov_wref(self):
        self.assertTrue(np.allclose(self.ge_cov_wref.error_rel, self.ref_err_val_rel, atol=1e-3))

    def test_convert_error_from_cor_rel_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_rel_wref.error, self.ref_err_val, atol=1e-3))

    def test_convert_error_rel_from_cor_wref(self):
        self.assertTrue(np.allclose(self.ge_cor_wref.error_rel, self.ref_err_val_rel, atol=1e-3))

    def test_raise_cov_from_cov_rel_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_rel_noref.cov_mat

    def test_raise_cov_rel_from_cov_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_noref.cov_mat_rel

    def test_raise_cov_from_cor_rel_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cor_rel_noref.cov_mat

    def test_raise_cov_rel_from_cor_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cor_noref.cov_mat_rel

    def test_raise_error_from_cov_rel_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_rel_noref.error

    def test_raise_error_rel_from_cov_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_noref.error_rel

    def test_raise_error_from_cor_rel_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cor_rel_noref.error

    def test_raise_error_rel_from_cor_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cor_noref.error_rel

    def test_raise_build_from_wrong_cor_mat_diagonal_not_ones(self):
        with self.assertRaises(ValueError):
            MatrixGaussianError(self.wrong_cor_mat_diagonal_not_ones, 'cor', err_val=self.ref_err_val, reference=self.ref_reference)

    def test_raise_build_from_wrong_cov_mat_asymm(self):
        with self.assertWarns(UserWarning):
            _err = MatrixGaussianError(self.wrong_cov_mat_asymm, 'cov', err_val=None,
                                reference=self.ref_reference)
            _err.check_cov_mat_symmetry()

    def test_raise_build_from_wrong_cor_mat_asymm(self):
        with self.assertWarns(UserWarning):
            _err = MatrixGaussianError(self.wrong_cor_mat_asymm, 'cor', err_val=self.ref_err_val,
                                reference=self.ref_reference)
            _err.check_cov_mat_symmetry()

    def test_raise_build_cov_mat_with_err_val(self):
        with self.assertRaises(ValueError):
            MatrixGaussianError(self.ref_cov_mat, 'cov', err_val=self.ref_err_val)

    def test_raise_build_cor_mat_without_err_val(self):
        with self.assertRaises(ValueError):
            MatrixGaussianError(self.ref_cor_mat, 'cor')

    def test_raise_build_unknown_mat_type(self):
        with self.assertRaises(ValueError):
            MatrixGaussianError(self.ref_cov_mat, 'DEADBEEF')

    def test_raise_attribute_errors(self):
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.error_cor
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.error_uncor
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.error_rel_cor
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.error_rel_uncor
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.cov_mat_cor
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.cov_mat_uncor
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.cov_mat_rel_cor
        with self.assertRaises(AttributeError):
            _ = self.ge_cov_wref.cov_mat_rel_uncor

    def test_raise_bad_input_shape(self):
        with self.assertRaises(ValueError):
            _ = MatrixGaussianError([1, 0], "cov")
        with self.assertRaises(ValueError):
            _ = MatrixGaussianError([[1, 0], [0, 1]], "cor", err_val=[[1, 2], [3, 4]])

class TestSimpleGaussianError(unittest.TestCase):

    def setUp(self):
        self.ref_reference = np.array([0.1, 0.2, 0.3])
        self.ref_error = np.array([0.1, 0.2, 0.1])
        self.ref_corr_coeff = 0.3

        # construct the reference absolute error vectors
        self.ref_error_uncor = self.ref_error * np.sqrt(1 - self.ref_corr_coeff)
        self.ref_error_cor = self.ref_error * np.sqrt(self.ref_corr_coeff)

        # construct the reference relative error vectors
        self.ref_error_rel = self.ref_error / self.ref_reference
        self.ref_error_rel_uncor = self.ref_error_rel * np.sqrt(1 - self.ref_corr_coeff)
        self.ref_error_rel_cor = self.ref_error_rel * np.sqrt(self.ref_corr_coeff)

        # construct the reference absolute covariance matrices
        self.ref_cov_mat_uncor = np.diag(self.ref_error_uncor ** 2)
        self.ref_cov_mat_cor = np.outer(self.ref_error_cor, self.ref_error_cor)

        self.ref_cov_mat = self.ref_cov_mat_uncor + self.ref_cov_mat_cor
        self.ref_cor_mat = self.ref_cov_mat / np.outer(self.ref_error, self.ref_error)
        self.ref_cov_mat_inverse = np.linalg.inv(self.ref_cov_mat)

        # construct the reference relative covariance matrices
        self.ref_cov_mat_rel_uncor = np.diag(self.ref_error_rel_uncor ** 2)
        self.ref_cov_mat_rel_cor = np.outer(self.ref_error_rel_cor, self.ref_error_rel_cor)

        self.ref_cov_mat_rel = self.ref_cov_mat_rel_uncor + self.ref_cov_mat_rel_cor
        self.ref_cov_mat_rel_inverse = np.linalg.inv(self.ref_cov_mat_rel)

        # -- errors with no reference

        self.sge_abs_noref = SimpleGaussianError(err_val=self.ref_error,
                                                 corr_coeff=self.ref_corr_coeff,
                                                 relative=False,
                                                 reference=None)
        self.sge_abs_wref = SimpleGaussianError(err_val=self.ref_error,
                                                corr_coeff=self.ref_corr_coeff,
                                                relative=False,
                                                reference=self.ref_reference)
        self.sge_rel_noref = SimpleGaussianError(err_val=self.ref_error_rel,
                                                 corr_coeff=self.ref_corr_coeff,
                                                 relative=True,
                                                 reference=None)
        self.sge_rel_wref = SimpleGaussianError(err_val=self.ref_error_rel,
                                                 corr_coeff=self.ref_corr_coeff,
                                                 relative=True,
                                                 reference=self.ref_reference)
        self.sge_abs_neg_ref = SimpleGaussianError(err_val=self.ref_error,
                                                   corr_coeff=self.ref_corr_coeff,
                                                   relative=False,
                                                   reference=-self.ref_reference)
        self.sge_rel_neg_ref = SimpleGaussianError(err_val=self.ref_error,
                                                   corr_coeff=self.ref_corr_coeff,
                                                   relative=False,
                                                   reference=-self.ref_reference)

    def test_constructor_raise(self):
        with self.assertRaises(ValueError):
            _ = SimpleGaussianError(err_val=1, corr_coeff=0.5)
        with self.assertRaises(ValueError):
            _ = SimpleGaussianError(err_val=[[1]], corr_coeff=0.5)
        with self.assertRaises(ValueError):
            _ = SimpleGaussianError(err_val=[1], corr_coeff=2.0)
        with self.assertRaises(ValueError):
            _ = SimpleGaussianError(err_val=[1], corr_coeff=-1.0)

    def test_set_error_raise(self):
        self.sge_abs_noref.error = 2.0
        with self.assertRaises(ValueError):
            self.sge_abs_noref.error = -1.0
        self.sge_abs_wref.error = 2.0
        with self.assertRaises(ValueError):
            self.sge_abs_wref.error = -1.0
        self.sge_rel_wref.error = 2.0
        with self.assertRaises(ValueError):
            self.sge_rel_wref.error = -1.0
        with self.assertRaises(AttributeError):
            self.sge_rel_noref.error = 2.0
        with self.assertRaises(ValueError):
            self.sge_rel_noref.error = -1.0

    def test_set_error_rel_raise(self):
        with self.assertRaises(AttributeError):
            self.sge_abs_noref.error_rel = 2.0
        with self.assertRaises(ValueError):
            self.sge_abs_noref.error_rel = -1.0
        self.sge_abs_wref.error_rel = 2.0
        with self.assertRaises(ValueError):
            self.sge_abs_wref.error_rel = -1.0
        self.sge_rel_wref.error_rel = 2.0
        with self.assertRaises(ValueError):
            self.sge_rel_wref.error_rel = -1.0
        self.sge_rel_noref.error_rel = 2.0

    # -- test errors with a reference

    # --- test covariance matrices

    def test_compare_cov_from_abs_wref(self):
        self.assertTrue(np.allclose(self.sge_abs_wref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_compare_cov_cor_uncor_from_abs_wref(self):
        self.assertTrue(np.allclose(
            self.sge_abs_wref.cov_mat_cor, self.ref_cov_mat_cor, atol=1e-3))
        self.assertTrue(np.allclose(
            self.sge_abs_wref.cov_mat_uncor, self.ref_cov_mat_uncor, atol=1e-3))

    def test_compare_cov_rel_from_abs_wref(self):
        self.assertTrue(np.allclose(self.sge_abs_wref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_compare_cov_rel_cor_uncor_from_abs_wref(self):
        self.assertTrue(np.allclose(
            self.sge_abs_wref.cov_mat_rel_cor, self.ref_cov_mat_rel_cor, atol=1e-3))
        self.assertTrue(np.allclose(
            self.sge_abs_wref.cov_mat_rel_uncor, self.ref_cov_mat_rel_uncor, atol=1e-3))

    def test_compare_cov_from_rel_wref(self):
        self.assertTrue(np.allclose(self.sge_rel_wref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_compare_cov_cor_uncor_from_rel_wref(self):
        self.assertTrue(np.allclose(
            self.sge_rel_wref.cov_mat_cor, self.ref_cov_mat_cor, atol=1e-3))
        self.assertTrue(np.allclose(
            self.sge_rel_wref.cov_mat_uncor, self.ref_cov_mat_uncor, atol=1e-3))
        
    def test_compare_cov_rel_from_rel_wref(self):
        self.assertTrue(np.allclose(self.sge_rel_wref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_compare_cov_rel_cor_uncor_from_rel_wref(self):
        self.assertTrue(np.allclose(
            self.sge_rel_wref.cov_mat_rel_uncor, self.ref_cov_mat_rel_uncor, atol=1e-3))
        self.assertTrue(np.allclose(
            self.sge_rel_wref.cov_mat_rel_cor, self.ref_cov_mat_rel_cor, atol=1e-3))

    # --- test error vectors
    def test_compare_error_from_abs_wref(self):
        self.assertTrue(np.allclose(self.sge_abs_wref.error, self.ref_error, atol=1e-3))

    def test_compare_error_cor_uncor_from_abs_wref(self):
        self.assertTrue(np.allclose(self.sge_abs_wref.error_cor, self.ref_error_cor, atol=1e-3))
        self.assertTrue(np.allclose(self.sge_abs_wref.error_uncor, self.ref_error_uncor, atol=1e-3))

    def test_compare_error_rel_from_abs_wref(self):
        self.assertTrue(np.allclose(self.sge_abs_wref.error_rel, self.ref_error_rel, atol=1e-3))

    def test_compare_error_rel_cor_uncor_from_abs_wref(self):
        self.assertTrue(np.allclose(self.sge_abs_wref.error_rel_cor, self.ref_error_rel_cor, atol=1e-3))
        self.assertTrue(np.allclose(self.sge_abs_wref.error_rel_uncor, self.ref_error_rel_uncor, atol=1e-3))

    def test_compare_error_from_rel_wref(self):
        self.assertTrue(np.allclose(self.sge_rel_wref.error, self.ref_error, atol=1e-3))

    def test_compare_error_cor_uncor_from_rel_wref(self):
        self.assertTrue(np.allclose(self.sge_rel_wref.error_cor, self.ref_error_cor, atol=1e-3))
        self.assertTrue(np.allclose(self.sge_rel_wref.error_uncor, self.ref_error_uncor, atol=1e-3))

    def test_compare_error_rel_from_rel_wref(self):
        self.assertTrue(np.allclose(self.sge_rel_wref.error_rel, self.ref_error_rel, atol=1e-3))

    def test_compare_error_rel_cor_uncor_from_rel_wref(self):
        self.assertTrue(np.allclose(self.sge_rel_wref.error_rel_cor, self.ref_error_rel_cor, atol=1e-3))
        self.assertTrue(np.allclose(self.sge_rel_wref.error_rel_uncor, self.ref_error_rel_uncor, atol=1e-3))

    def test_negative_reference(self):
        self.assertTrue(np.allclose(self.sge_abs_neg_ref.error, self.ref_error))
        self.assertTrue(np.allclose(self.sge_abs_neg_ref.error_rel, self.ref_error_rel))
        self.assertTrue(np.allclose(self.sge_rel_neg_ref.error, self.ref_error))
        self.assertTrue(np.allclose(self.sge_rel_neg_ref.error_rel, self.ref_error_rel))

    # -- test errors without a reference

    def test_compare_cov_from_abs_noref(self):
        self.assertTrue(np.allclose(self.sge_abs_noref.cov_mat, self.ref_cov_mat, atol=1e-3))

    def test_compare_cov_rel_from_rel_noref(self):
        self.assertTrue(np.allclose(
            self.sge_rel_noref.cov_mat_rel, self.ref_cov_mat_rel, atol=1e-3))

    def test_compare_error_from_abs_noref(self):
        self.assertTrue(np.allclose(self.sge_abs_noref.error, self.ref_error, atol=1e-3))

    def test_compare_error_rel_from_rel_noref(self):
        self.assertTrue(np.allclose(self.sge_rel_noref.error_rel, self.ref_error_rel, atol=1e-3))

    def test_raise_cov_from_rel_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.sge_rel_noref.cov_mat

    def test_raise_cov_rel_from_abs_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.sge_abs_noref.cov_mat_rel

    def test_raise_error_from_rel_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.sge_rel_noref.error

    def test_raise_error_rel_from_abs_noref(self):
        with self.assertRaises(AttributeError):
            _ = self.sge_abs_noref.error_rel

    def test_compare_cov_mat_inverse(self):
        self.assertTrue(np.allclose(self.sge_abs_noref.cov_mat_inverse, self.ref_cov_mat_inverse))
        self.assertTrue(np.allclose(self.sge_abs_wref.cov_mat_inverse, self.ref_cov_mat_inverse))
        with self.assertRaises(AttributeError):
            self.assertTrue(np.allclose(
                self.sge_rel_noref.cov_mat_inverse, self.ref_cov_mat_inverse))
        self.assertTrue(np.allclose(self.sge_rel_wref.cov_mat_inverse, self.ref_cov_mat_inverse))

    def test_compare_cov_mat_rel_inverse(self):
        with self.assertRaises(AttributeError):
            self.assertTrue(np.allclose(
                self.sge_abs_noref.cov_mat_rel_inverse, self.ref_cov_mat_rel_inverse))
        self.assertTrue(np.allclose(
            self.sge_abs_wref.cov_mat_rel_inverse, self.ref_cov_mat_rel_inverse))
        self.assertTrue(np.allclose(
            self.sge_rel_noref.cov_mat_rel_inverse, self.ref_cov_mat_rel_inverse))
        self.assertTrue(np.allclose(
            self.sge_rel_wref.cov_mat_rel_inverse, self.ref_cov_mat_rel_inverse))

    def test_compare_cor_mat(self):
        self.assertTrue(np.allclose(self.sge_abs_noref.cor_mat, self.ref_cor_mat))
        self.assertTrue(np.allclose(self.sge_abs_wref.cor_mat, self.ref_cor_mat))
        self.assertTrue(np.allclose(self.sge_rel_noref.cor_mat, self.ref_cor_mat))
        self.assertTrue(np.allclose(self.sge_rel_wref.cor_mat, self.ref_cor_mat))



import numpy as np
from types import FunctionType

from .._base import CostFunction
from ..util import function_library

__all__ = [
    "UnbinnedCostFunction_NegLogLikelihood",
]


class UnbinnedCostFunction_NegLogLikelihood(CostFunction):
    def __init__(self):
        r"""
        Built-in negative log-likelihood cost function for *Unbinned* data.

        When using an unbinned dataset, the negative log-likelihood is the best method to fit a probability density
        funtion *pdf* to the density of the datapoints

        In general, a negative log-likelihood cost function is defined as the double negative logarithm of the
        product of the individual likelihoods of the data points.
        """
        super(UnbinnedCostFunction_NegLogLikelihood, self).__init__(
            cost_function=self.nll, add_determinant_cost=False)
        self._needs_errors = False
        self._formatter.latex_name = "-2\\ln\\mathcal{L}"
        self._formatter.name = "nll"
        self._formatter.description = "negative log-likelihood"

    # model is the pdf already evaluated at all x-points with the given params, as far as I understand.
    # so there's only need to evaluate the model in the nll calculations?
    @staticmethod
    def nll(model):
        _total_log_likelihood = np.sum(np.log(model))
        # guard against returning NaN
        if np.isnan(_total_log_likelihood):
            return np.inf
        return -2.0 * _total_log_likelihood


STRING_TO_COST_FUNCTION = {
    'nll': (UnbinnedCostFunction_NegLogLikelihood, {}),
    'negloglikelihood': (UnbinnedCostFunction_NegLogLikelihood, {}),
    'neg_log_likelihood': (UnbinnedCostFunction_NegLogLikelihood, {}),
}



import numpy as np
import unittest
from scipy.special import factorial

from kafe2.core.constraint import GaussianSimpleParameterConstraint, \
    GaussianMatrixParameterConstraint
from kafe2.fit._base.cost import *
from kafe2.fit.histogram.cost import *
from kafe2.fit.indexed.cost import *
from kafe2.fit.xy.cost import *


class TestCostBuiltin(unittest.TestCase):
    CHI2_COST_FUNCTION = CostFunction_Chi2
    NLL_COST_FUNCTION = CostFunction_NegLogLikelihood
    GAUSS_APPROXIMATION_COST_FUNCTION = CostFunction_GaussApproximation

    def setUp(self):
        self._data_chi2 = np.array([-0.5, 2.1, 8.9])
        self._model_chi2 = np.array([5.7, 8.4, -2.3])
        self._data_poisson = np.array([0.0, 2.0, 9.0])
        self._model_poisson = np.array([5.7, 8.4, 2.3])
        self._res = self._data_chi2 - self._model_chi2
        self._cov_mat = np.array([
            [1.0, 0.1, 0.2],
            [0.1, 2.0, 0.3],
            [0.2, 0.3, 3.0]
        ])
        self._cov_mat_cholesky = np.linalg.cholesky(self._cov_mat)
        self._cov_mat_inv = np.linalg.inv(self._cov_mat)
        self._pointwise_errors = np.sqrt(np.diag(self._cov_mat))
        self._data_poisson_large = np.array([100000, 200000, 300000])
        self._model_poisson_large = np.array([99990, 200010, 300000])

        self._cov_mat_log_det = np.log(np.linalg.det(self._cov_mat))
        self._total_error_squared_log_sum = 2.0 * np.sum(np.log(self._pointwise_errors))
        self._cost_chi2_cov_mat = self._res.dot(
            self._cov_mat_inv).dot(self._res) + self._cov_mat_log_det
        self._cost_chi2_pointwise = np.sum(
            (self._res / self._pointwise_errors) ** 2) + self._total_error_squared_log_sum
        self._cost_chi2_no_errors = np.sum(self._res ** 2)
        self._cost_nll_gaussian = self._cost_chi2_pointwise + 2.0 * np.sum(np.log(
            np.sqrt((2.0 * np.pi)) * self._pointwise_errors)) - self._total_error_squared_log_sum
        self._cost_nll_poisson = -2.0 * np.sum(
            np.log(self._model_poisson ** self._data_poisson)
            - np.log(factorial(self._data_poisson)) - self._model_poisson)
        self._cost_nllr_poisson = self._cost_nll_poisson + 2.0 * np.sum(
            np.log(self._data_poisson ** self._data_poisson)
            - np.log(factorial(self._data_poisson)) - self._data_poisson)
        self._par_vals = np.array([12.3, 0.001, -1.9])
        self._simple_constraint = GaussianSimpleParameterConstraint(
            index=0, value=10.0, uncertainty=2.5)
        self._matrix_constraint = GaussianMatrixParameterConstraint(
            indices=(0, 1, 2),
            values=(1.0, 2.0, 3.0),
            matrix=[
                [1.5, 0.1, 0.1],
                [0.1, 2.2, 0.1],
                [0.1, 0.1, 0.3]
            ]
        )
        self. _par_constraints = [self._simple_constraint, self._matrix_constraint]
        self._par_cost = self._simple_constraint.cost(self._par_vals) \
            + self._matrix_constraint.cost(self._par_vals)

    def test_chi2_no_errors(self):
        self.assertAlmostEqual(
            self._cost_chi2_no_errors,
            self.CHI2_COST_FUNCTION(errors_to_use=None, add_determinant_cost=False)
            (self._data_chi2, self._model_chi2, None, None))
        self.assertAlmostEqual(
            self._cost_chi2_no_errors + self._par_cost,
            self.CHI2_COST_FUNCTION(errors_to_use=None, add_determinant_cost=False)
            (self._data_chi2, self._model_chi2, self._par_vals, self._par_constraints))

    def test_chi2_pointwise(self):
        self.assertAlmostEqual(
            self._cost_chi2_pointwise,
            self.CHI2_COST_FUNCTION(errors_to_use='pointwise')
            (self._data_chi2, self._model_chi2, self._pointwise_errors, None, None,
             self._total_error_squared_log_sum))
        self.assertAlmostEqual(
            self._cost_chi2_pointwise + self._par_cost,
            self.CHI2_COST_FUNCTION(errors_to_use='pointwise')
            (self._data_chi2, self._model_chi2, self._pointwise_errors,
             self._par_vals, self._par_constraints, self._total_error_squared_log_sum))

    def test_chi2_cov_mat(self):
        self.assertAlmostEqual(
            self._cost_chi2_cov_mat,
            self.CHI2_COST_FUNCTION(errors_to_use='covariance')
            (self._data_chi2, self._model_chi2, self._cov_mat_cholesky, None, None,
             self._cov_mat_log_det))
        self.assertAlmostEqual(
            self._cost_chi2_cov_mat + self._par_cost,
            self.CHI2_COST_FUNCTION(errors_to_use='covariance')
            (self._data_chi2, self._model_chi2, self._cov_mat_cholesky,
             self._par_vals, self._par_constraints, self._cov_mat_log_det))

    def test_nll_gaussian(self):
        self.assertAlmostEqual(
            self._cost_nll_gaussian,
            self.NLL_COST_FUNCTION(data_point_distribution='gaussian')
            (self._data_chi2, self._model_chi2, self._pointwise_errors, None, None))
        self.assertAlmostEqual(
            self._cost_nll_gaussian + self._par_cost,
            self.NLL_COST_FUNCTION(data_point_distribution='gaussian')
            (self._data_chi2, self._model_chi2, self._pointwise_errors,
             self._par_vals, self._par_constraints))

    def test_nll_poisson(self):
        self.assertAlmostEqual(
            self._cost_nll_poisson,
            self.NLL_COST_FUNCTION(data_point_distribution='poisson')
            (self._data_poisson, self._model_poisson, None, None))
        self.assertAlmostEqual(
            self._cost_nll_poisson + self._par_cost,
            self.NLL_COST_FUNCTION(data_point_distribution='poisson')
            (self._data_poisson, self._model_poisson, self._par_vals, self._par_constraints))

    def test_nllr_gaussian(self):
        self.assertAlmostEqual(
            self._cost_chi2_pointwise - self._total_error_squared_log_sum,
            self.NLL_COST_FUNCTION(data_point_distribution='gaussian', ratio=True)
            (self._data_chi2, self._model_chi2, self._pointwise_errors, None, None))
        self.assertAlmostEqual(
            self._cost_chi2_pointwise + self._par_cost - self._total_error_squared_log_sum,
            self.NLL_COST_FUNCTION(data_point_distribution='gaussian', ratio=True)
            (self._data_chi2, self._model_chi2, self._pointwise_errors,
             self._par_vals, self._par_constraints))

    def test_nllr_poisson(self):
        self.assertAlmostEqual(
            self._cost_nllr_poisson,
            self.NLL_COST_FUNCTION(data_point_distribution='poisson', ratio=True)
            (self._data_poisson, self._model_poisson, None, None))
        self.assertAlmostEqual(
            self._cost_nllr_poisson + self._par_cost,
            self.NLL_COST_FUNCTION(data_point_distribution='poisson', ratio=True)
            (self._data_poisson, self._model_poisson, self._par_vals, self._par_constraints))

    def test_gauss_approximation(self):
        self.assertAlmostEqual(
            self.NLL_COST_FUNCTION(data_point_distribution='poisson')
            (self._data_poisson_large, self._model_poisson_large, None, None),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use='covariance')
            (self._data_poisson_large, self._model_poisson_large,
             np.zeros_like(self._cov_mat), None, None) + 3 * np.log(2 * np.pi),
            places=3
        )
        self.assertAlmostEqual(
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use='covariance')
            (self._data_poisson, self._model_poisson,
             np.zeros_like(self._cov_mat), None, None),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use='pointwise')
            (self._data_poisson, self._model_poisson,
             np.zeros_like(self._pointwise_errors), None, None),
        )
        self.assertAlmostEqual(
            self.CHI2_COST_FUNCTION(errors_to_use='covariance')
            (self._data_chi2, np.zeros_like(self._model_chi2), self._cov_mat_cholesky, None, None,
             self._cov_mat_log_det),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use='covariance')
            (self._data_chi2, np.zeros_like(self._model_chi2), self._cov_mat, None, None),
        )
        self.assertAlmostEqual(
            self.CHI2_COST_FUNCTION(errors_to_use='pointwise')
            (self._data_chi2, np.zeros_like(self._model_chi2), self._pointwise_errors, None, None,
             self._total_error_squared_log_sum),
            self.GAUSS_APPROXIMATION_COST_FUNCTION(errors_to_use='pointwise')
            (self._data_chi2, np.zeros_like(self._model_chi2), self._pointwise_errors, None, None),
        )


    def test_chi2_raise(self):
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="XYZ")
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2, np.ones(10), self._cov_mat_cholesky, None, None,
                self._cov_mat_log_det)
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="covariance", fallback_on_singular=False)(
                self._data_chi2, self._model_chi2, None, None, None, self._cov_mat_log_det)
        with self.assertRaises(ValueError):
            self.CHI2_COST_FUNCTION(errors_to_use="pointwise", fallback_on_singular=False)(
                self._data_chi2, self._model_chi2, np.arange(len(self._pointwise_errors)),
                None, None, self._cov_mat_log_det)

    def test_nll_raise(self):
        with self.assertRaises(ValueError):
            self.NLL_COST_FUNCTION(data_point_distribution="yes")

    def test_inf_cost(self):
        self.assertEqual(
            np.inf,
            self.CHI2_COST_FUNCTION(errors_to_use="covariance")(
                self._data_chi2, np.nan * np.ones_like(self._model_chi2), self._cov_mat_cholesky,
                None, None, self._cov_mat_log_det)
        )
        self.assertEqual(
            np.inf,
            self.CHI2_COST_FUNCTION(errors_to_use="pointwise")(
                self._data_chi2, np.nan * np.ones_like(self._model_chi2), self._pointwise_errors,
                None, None, self._cov_mat_log_det)
        )
        self.assertEqual(
            np.inf,
            self.CHI2_COST_FUNCTION(errors_to_use=None)(
                self._data_chi2, np.nan * np.ones_like(self._model_chi2), None, None,
                self._cov_mat_log_det)
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("poisson", ratio=False)(
                self._data_poisson, -self._model_poisson, None, None)
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("poisson", ratio=True)(
                self._data_poisson, -self._model_poisson, None, None)
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("gaussian", ratio=False)(
                self._data_chi2, self._model_chi2, -self._pointwise_errors, None, None)
        )
        self.assertEqual(
            np.inf,
            self.NLL_COST_FUNCTION("gaussian", ratio=True)(
                self._data_chi2, self._model_chi2, -self._pointwise_errors, None, None)
        )


class TestCostBuiltinHist(TestCostBuiltin):

    CHI2_COST_FUNCTION = HistCostFunction_Chi2
    NLL_COST_FUNCTION = HistCostFunction_NegLogLikelihood


class TestCostBuiltinIndexed(TestCostBuiltin):

    CHI2_COST_FUNCTION = IndexedCostFunction_Chi2
    NLL_COST_FUNCTION = IndexedCostFunction_NegLogLikelihood


class TestCostBuiltinXY(TestCostBuiltin):

    CHI2_COST_FUNCTION = XYCostFunction_Chi2
    NLL_COST_FUNCTION = XYCostFunction_NegLogLikelihood


class TestCostUserDefined(unittest.TestCase):
    def setUp(self):
        def my_cost(a, b, c, d):
            return a ** 2 + (b + c) ** 2 + (d - 1) ** 2

        def my_cost_varargs(*args):
            return args[0] ** 2 + (args[1] + args[2]) ** 2 + (args[3] - 1) ** 2

        self._ref_par_vals = [1, 2, 5, 7]
        self._ref_cost = my_cost(*self._ref_par_vals)
        self._constraint = GaussianSimpleParameterConstraint(index=0, value=0, uncertainty=1)
        self._matrix_constraint = GaussianMatrixParameterConstraint(
            indices=[1], values=[0], matrix=[[1]])
        self._constraint_cost = self._ref_par_vals[0] ** 2 + self._ref_par_vals[1] ** 2
        self._ref_par_vals_constraints = self._ref_par_vals + [
            self._ref_par_vals, [self._constraint, self._matrix_constraint]]

        self._cost_func = CostFunction(
            my_cost, arg_names=None, add_constraint_cost=False, add_determinant_cost=False)
        self._cost_func_constraints = CostFunction(
            my_cost, arg_names=None, add_constraint_cost=True, add_determinant_cost=False)
        self._cost_func_varargs = CostFunction(
            my_cost_varargs, arg_names=["a", "b", "c", "d"], add_constraint_cost=False,
            add_determinant_cost=False)
        self._cost_func_varargs_constraints = CostFunction(
            my_cost_varargs, arg_names=["a", "b", "c", "d"], add_constraint_cost=True,
            add_determinant_cost=False)

    def test_properties(self):
        self.assertEqual(self._cost_func.name, "my_cost")
        self.assertEqual(self._cost_func.arg_names, ["a", "b", "c", "d"])
        self.assertEqual(self._cost_func_varargs.name, "my_cost_varargs")
        self.assertEqual(self._cost_func.arg_names, ["a", "b", "c", "d"])

    def test_validate_raise(self):
        def _cost_args(*args):
            return np.sum(args)

        def _cost_kwargs(**kwargs):
            return np.sum(list(kwargs.values()))

        def _cost_bad_arg_name(cost):
            return cost

        _ = CostFunction(_cost_args, arg_names=["a", "b"])
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_args)
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_kwargs)
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_kwargs, arg_names=["a", "b"])
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_bad_arg_name)
        _ = CostFunction(_cost_bad_arg_name, arg_names=["a"])
        with self.assertRaises(ValueError):
            _ = CostFunction(_cost_args, arg_names=["cost"])

    def test_compare_cost(self):
        self.assertEqual(
            self._ref_cost, self._cost_func(*self._ref_par_vals)
        )
        self.assertEqual(
            self._ref_cost, self._cost_func_constraints(*(self._ref_par_vals + [None, None]))
        )
        self.assertEqual(
            self._ref_cost,
            self._cost_func_constraints(*(self._ref_par_vals + [self._ref_par_vals, None]))
        )
        self.assertEqual(
            self._ref_cost + self._constraint_cost,
            self._cost_func_constraints(*self._ref_par_vals_constraints)
        )

    def test_compare_cost_varargs(self):
        self.assertEqual(
            self._ref_cost, self._cost_func_varargs(*self._ref_par_vals)
        )
        self.assertEqual(
            self._ref_cost, self._cost_func_varargs_constraints(*(
                    self._ref_par_vals + [None, None]))
        )
        self.assertEqual(
            self._ref_cost,
            self._cost_func_varargs_constraints(*(
                    self._ref_par_vals + [self._ref_par_vals, None]))
        )
        self.assertEqual(
            self._ref_cost + self._constraint_cost,
            self._cost_func_varargs_constraints(*self._ref_par_vals_constraints)
        )




from collections import OrderedDict
from copy import deepcopy

import numpy as np

from ...tools import print_dict_as_table
from .._base import FitBase, DataContainerBase
from .container import IndexedContainer
from .._base.cost import CostFunction_Chi2
from .model import IndexedParametricModel, IndexedModelFunction
from .plot import IndexedPlotAdapter
from ..util import collect


__all__ = ['IndexedFit']


class IndexedFit(FitBase):
    CONTAINER_TYPE = IndexedContainer
    MODEL_TYPE = IndexedParametricModel
    MODEL_FUNCTION_TYPE = IndexedModelFunction
    PLOT_ADAPTER_TYPE = IndexedPlotAdapter
    RESERVED_NODE_NAMES = {'data', 'model', 'cost',
                          'data_error', 'model_error', 'total_error',
                          'data_cov_mat', 'model_cov_mat', 'total_cov_mat',
                          'data_cor_mat', 'model_cor_mat', 'total_cor_mat'}
    _BASIC_ERROR_NAMES = {'data_error', 'model_error', 'data_cov_mat', 'model_cov_mat'}

    def __init__(self,
                 data,
                 model_function,
                 cost_function="chi2",
                 minimizer=None,
                 minimizer_kwargs=None,
                 dynamic_error_algorithm="nonlinear"):
        """
        Construct a fit of a model to a series of indexed measurements.

        :param data: the measurement values
        :type data: iterable of float
        :param model_function: the model function
        :type model_function: :py:class:`~kafe2.fit.indexed.IndexedModelFunction` or unwrapped native Python function
        :param cost_function: the cost function
        :type cost_function: :py:class:`~kafe2.fit._base.CostFunctionBase`-derived or unwrapped native Python function
        :param minimizer: the minimizer to use for fitting.
        :type minimizer: None, "iminuit", "tminuit", or "scipy".
        :param minimizer_kwargs: dictionary with kwargs for the minimizer.
        :type minimizer_kwargs: dict
        """
        super(IndexedFit, self).__init__(
            data=data, model_function=model_function, cost_function=cost_function,
            minimizer=minimizer, minimizer_kwargs=minimizer_kwargs,
            dynamic_error_algorithm=dynamic_error_algorithm
        )

    # -- private methods

    def _init_nexus(self):
        super(IndexedFit, self)._init_nexus()

        self._nexus.add_alias("y_data", alias_for="data")
        self._nexus.add_alias("y_model", alias_for="model")
        self._nexus.add_alias("y_total_cov_mat", alias_for="total_cov_mat")

        self._nexus.add_dependency(
            'model',
            depends_on=(
                'parameter_values'
            )
        )

    def _report_data(self, output_stream, indent, indentation_level):
        output_stream.write(indent * indentation_level + '########\n')
        output_stream.write(indent * indentation_level + '# Data #\n')
        output_stream.write(indent * indentation_level + '########\n\n')

        _data_table_dict = OrderedDict()
        _data_table_dict['Index'] = range(self.data_size)
        _data_table_dict['Data'] = self.data
        if self.has_data_errors:
            _data_table_dict['Data Error'] = self.data_error
            _data_table_dict['Data Correlation Matrix'] = self.data_cor_mat

        print_dict_as_table(_data_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write('\n')

    def _report_model(self, output_stream, indent, indentation_level):
        # call base method to show header and model function
        super(IndexedFit, self)._report_model(output_stream, indent, indentation_level)
        _model_table_dict = OrderedDict()
        _model_table_dict['Index'] = range(self.data_size)
        _model_table_dict['Model'] = self.model
        if self.has_model_errors:
            _model_table_dict['Model Error'] = self.model_error
            _model_table_dict['Model Correlation Matrix'] = self.model_cor_mat

        print_dict_as_table(_model_table_dict, output_stream=output_stream, indent_level=indentation_level + 1)
        output_stream.write('\n')

        if self._param_model.get_matching_errors({"relative": True}):
            output_stream.write(indent * (indentation_level + 1))
            output_stream.write(
                "Model covariance matrix was calculated dynamically relative to y model values.\n"
            )
            output_stream.write("\n")

    # -- public properties

    def _set_new_data(self, new_data):
        if isinstance(new_data, self.CONTAINER_TYPE):
            self._data_container = deepcopy(new_data)
        elif isinstance(new_data, DataContainerBase):
            raise TypeError("Incompatible container type '%s' (expected '%s')"
                                      % (type(new_data), self.CONTAINER_TYPE))
        else:
            self._data_container = IndexedContainer(new_data, dtype=float)
        self._data_container._on_error_change_callback = self._on_error_change

        self._nexus.get('data').mark_for_update()

    def _set_new_parametric_model(self):
        self._param_model = IndexedParametricModel(
            self._model_function,
            self.parameter_values,
            shape_like=self.data
        )

    def _get_default_x_ticks(self):
        if self.data_size <= 10:
            return np.arange(self.data_size)
        else:
            return []

    # -- public properties

    @property
    def model(self):
        """array of model predictions for the data points"""
        self._param_model.parameters = self.parameter_values  # this is lazy, so just do it
        return self._param_model.data



#!/usr/bin/env python

import unittest
import sys
import os

from setuptools import find_packages, setup

def discover_kafe_tests():
    _tl = unittest.TestLoader()
    _ts = _tl.discover('kafe2/test', 'test_*.py')
    return _ts

def read_local(filename):
    _path = os.path.join(os.path.dirname(__file__), filename)
    if os.path.exists(_path):
        return open(_path).read()
    return ""

def get_version():
    '''get kafe2 version without importing the whole package'''
    _tmp_locals = {}
    exec(read_local("kafe2/_version_info.py"), _tmp_locals)
    return _tmp_locals['__version__']

setup(
    name='kafe2',
    version=get_version(),
    description='Karlsruhe Fit Environment 2: a package for fitting and elementary data analysis',
    long_description=read_local('README.rst'),
    long_description_content_type="text/x-rst",
    author='Daniel Savoiu',
    author_email='daniel.savoiu@cern.ch',
    url='http://github.com/dsavoiu/kafe2',
    packages=find_packages(),
    package_data={'kafe2': ['config/*.conf', 'config/*.yaml', 'config/*.yml', 'fit/tools/kafe2go']},
    scripts=['kafe2/fit/tools/kafe2go.py', 'kafe2/fit/tools/kafe2go'],
    test_suite='setup.discover_kafe_tests',
    keywords=("kafe2 kit karlsruhe data analysis lab laboratory practical courses "
              "education university students physics fitting minimization minimisation "
              "regression parametric parameter estimation optimization optimisation"),
    license='GPL3',
    #TODO requirement versions
    install_requires=[
        'NumPy',
        'Numdifftools',
        'Scipy',
        'tabulate',
        'matplotlib',
        'PyYaml',
        'six',
        'funcsigs',
        'SymPy',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Visualization',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
)



from .._base.cost import CostFunction, CostFunction_Chi2, CostFunction_NegLogLikelihood, \
    CostFunction_GaussApproximation


__all__ = [
    "IndexedCostFunction",
    "IndexedCostFunction_Chi2",
    "IndexedCostFunction_NegLogLikelihood",
    "IndexedCostFunction_GaussApproximation"
]


class IndexedCostFunction(CostFunction):
    pass


class IndexedCostFunction_Chi2(CostFunction_Chi2):
    pass


class IndexedCostFunction_NegLogLikelihood(CostFunction_NegLogLikelihood):
    pass


class IndexedCostFunction_GaussApproximation(CostFunction_GaussApproximation):
    pass



import abc
import unittest
import numpy as np
import six

from kafe2.core.minimizers import AVAILABLE_MINIMIZERS

from kafe2.config import kc

from kafe2.fit import XYFit
from kafe2.fit.xy.cost import XYCostFunction_Chi2

from kafe2.test.fit.test_fit import AbstractTestFit


def simple_chi2(y_data, y_model):
    return np.sum((y_data - y_model)**2)

def simple_chi2_explicit_model_name(y_data, simple_xy_model):
    return np.sum((y_data - simple_xy_model)**2)

def simple_xy_model(x, a=1.1, b=2.2, c=3.3):
    return a * x ** 2 + b * x + c

def line_xy_model(x, a=3.0, b=0.0):
    return a * x + b


def analytic_solution(des_mat, cov_mat_inv, y_data):
    return (
        np.squeeze(np.asarray(np.linalg.inv(
            des_mat.T.dot(cov_mat_inv).dot(des_mat)
        ).dot(des_mat.T)
         .dot(cov_mat_inv)
         .dot(y_data))))


class TestXYFitBasicInterface(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        self._n_points = 10

        # "jitter" for data smearing
        self._y_jitter = np.array([-0.3193475 , -1.2404198 , -1.4906926 , -0.78832446,
                                   -1.7638106,   0.36664261,  0.49433821,  0.0719646,
                                    1.95670326,  0.31200215])
        assert len(self._y_jitter) == self._n_points

        # reference initial values
        self._ref_initial_pars = np.array([1.1, 2.2, 3.3])
        self._ref_x = np.arange(self._n_points)
        self._ref_initial_y_model = simple_xy_model(self._ref_x, *self._ref_initial_pars)
        self._ref_initial_xy_model = np.array([self._ref_x, self._ref_initial_y_model])

        # fit data
        self._ref_y_data = self._ref_initial_y_model + self._y_jitter
        self._ref_xy_data = np.array([self._ref_x, self._ref_y_data])

        self._linear_design_matrix = np.array([
            self._ref_x ** 2,
            self._ref_x,
            np.ones_like(self._ref_x)
        ]).T

        # pre-fit cost value
        self._ref_initial_cost = simple_chi2(
            y_data=self._ref_y_data,
            y_model=self._ref_initial_y_model,
        )

        # reference matrices/errors
        self._ref_error = np.ones(self._n_points)
        self._ref_matrix_eye = np.eye(self._n_points)

        # reference fit result values
        #self._nominal_fit_result_pars = (1.1351433, 2.13736919, 2.33346549)
        self._nominal_fit_result_pars = analytic_solution(
            self._linear_design_matrix,
            np.linalg.inv(self._ref_matrix_eye),
            self._ref_y_data,
        )

        self._nominal_fit_result_y_model = simple_xy_model(self._ref_x, *self._nominal_fit_result_pars)
        self._nominal_fit_result_xy_model = np.array([self._ref_x, self._nominal_fit_result_y_model])
        self._nominal_fit_result_cost = simple_chi2(
            y_data=self._ref_y_data,
            y_model=self._nominal_fit_result_y_model,
        )

        # helper dict with all reference property values
        self._ref_prop_dict = dict(
            did_fit=False,
            model_count=1,

            parameter_values=self._ref_initial_pars,
            parameter_names=('a', 'b', 'c'),
            cost_function_value=self._ref_initial_cost,

            x_data=self._ref_x,
            y_data=self._ref_y_data,
            data=self._ref_xy_data,
            x_data_error=self._ref_error * 0,
            y_data_error=self._ref_error,
            data_error=self._ref_error,
            x_data_cov_mat=self._ref_matrix_eye * 0,
            y_data_cov_mat=self._ref_matrix_eye,
            data_cov_mat=self._ref_matrix_eye,
            x_data_cov_mat_inverse=None,
            y_data_cov_mat_inverse=self._ref_matrix_eye,
            data_cov_mat_inverse=self._ref_matrix_eye,
            x_data_cor_mat=self._ref_matrix_eye * np.nan,
            y_data_cor_mat=self._ref_matrix_eye,
            data_cor_mat=self._ref_matrix_eye,

            x_model=self._ref_x,
            y_model=self._ref_initial_y_model,
            model=self._ref_initial_xy_model,
            x_model_error=self._ref_error * 0,
            y_model_error=self._ref_error * 0,
            model_error=self._ref_error * 0,
            x_model_cov_mat=self._ref_matrix_eye * 0,
            y_model_cov_mat=self._ref_matrix_eye * 0,
            model_cov_mat=self._ref_matrix_eye * 0,
            x_model_cov_mat_inverse=None,
            y_model_cov_mat_inverse=None,
            model_cov_mat_inverse=None,
            x_model_cor_mat=self._ref_matrix_eye * np.nan,
            y_model_cor_mat=self._ref_matrix_eye * np.nan,
            model_cor_mat=self._ref_matrix_eye * np.nan,

            x_total_error=self._ref_error * 0,
            y_total_error=self._ref_error,
            total_error=self._ref_error,
            x_total_cov_mat=self._ref_matrix_eye * 0,
            y_total_cov_mat=self._ref_matrix_eye,
            total_cov_mat=self._ref_matrix_eye,
            x_total_cov_mat_inverse=None,
            y_total_cov_mat_inverse=self._ref_matrix_eye,
            total_cov_mat_inverse=self._ref_matrix_eye,
            x_total_cor_mat=self._ref_matrix_eye * np.nan,
            y_total_cor_mat=self._ref_matrix_eye,
            total_cor_mat=self._ref_matrix_eye,
        )

    def _get_fit(
            self, model_function=None, cost_function=None, errors=None,
            dynamic_error_algorithm=None):
        '''convenience'''
        model_function = model_function or simple_xy_model
        # TODO: fix default
        cost_function = cost_function or XYCostFunction_Chi2(
            axes_to_use='xy', errors_to_use='covariance')
        errors = errors or [dict(axis='y', err_val=1.0)]
        dynamic_error_algorithm = dynamic_error_algorithm or "nonlinear"

        _fit = XYFit(
            xy_data=self._ref_xy_data,
            model_function=model_function,
            cost_function=cost_function,
            minimizer=self.MINIMIZER,
            dynamic_error_algorithm=dynamic_error_algorithm
        )
        for _err in errors:
            _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            'default': self._get_fit(),
            'explicit': self._get_fit(cost_function=simple_chi2),
            'explicit_model': self._get_fit(cost_function=simple_chi2_explicit_model_name),
            'relative_errors_data': self._get_fit(errors=[
                dict(axis="y", err_val=1.0/self._ref_y_data, relative=True, reference="data")
            ])
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                y_model=self._nominal_fit_result_y_model,
                model=self._nominal_fit_result_xy_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_set_all_parameter_values(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                x_data=self._ref_x,
                x_model=self._ref_x,
                y_data=self._ref_y_data,
                y_model=self._nominal_fit_result_y_model,
                data=self._ref_xy_data,
                model=self._nominal_fit_result_xy_model,
                parameter_names=('a', 'b', 'c'),
                did_fit=False,
                model_count=1,
                cost_function_value=self._nominal_fit_result_cost
            ),
            # set parameters to their final values by hand
            call_before_fit=lambda f: f.set_all_parameter_values(
                self._nominal_fit_result_pars),
        )

    def test_set_all_parameter_values_wrong_number_raise(self):
        with self.assertRaises(ValueError):
            self._get_fit().set_all_parameter_values((1,))
        with self.assertRaises(ValueError):
            self._get_fit().set_all_parameter_values((1,2,3,4,5))

    def test_parameter_defaults(self):
        def dummy_model(x, a, b, c):
            return x

        _fit = self._get_fit(model_function=dummy_model)

        self._assert_fit_properties(
            _fit,
            dict(
                parameter_values=np.array([1, 1, 1]),
            )
        )

    def test_parameter_partial_defaults(self):
        def dummy_model(x, a, b, c=3.3):
            return x

        _fit = self._get_fit(model_function=dummy_model)

        self._assert_fit_properties(
            _fit,
            dict(
                parameter_values=np.array([1, 1, 3.3]),
            )
        )

    def test_add_parameter_constraint(self):
        _fit = self._get_fit()

        _fit.add_parameter_constraint('c', 1.0, 1.0)
        _constraint_cost = (self._ref_initial_pars[2] - 1.0)**2

        self._assert_fit_properties(
            _fit,
            dict(
                cost_function_value=np.float64(
                    self._ref_initial_cost + _constraint_cost
                )
            )
        )

    def test_add_matrix_parameter_constraint(self):
        _fit = self._get_fit()

        _fit.add_matrix_parameter_constraint(('a', 'c'), np.ones(2), np.eye(2))
        _constraint_cost = (
            (self._ref_initial_pars[0] - 1.0)**2 +
            (self._ref_initial_pars[2] - 1.0)**2
        )

        self._assert_fit_properties(
            _fit,
            dict(
                cost_function_value=np.float64(
                    self._ref_initial_cost + _constraint_cost
                )
            )
        )

    def test_eval_model_function(self):
        _fit = self._get_fit()
        self.assertTrue(np.allclose(
            _fit.eval_model_function(self._ref_x),
            self._ref_initial_y_model
        ))

    def test_update_data(self):
        _fit = self._get_fit()

        _fit.data = np.array([self._ref_x, self._ref_y_data * 2])
        _fit.add_error('y', err_val=1.0)

        self._assert_fit_properties(
            _fit,
            dict(
                data=np.array([self._ref_x, self._ref_y_data * 2]),
                x_data=self._ref_x,
                y_data=self._ref_y_data * 2,
            )
        )

        _fit.do_fit()
        _new_estimates = np.array(self._nominal_fit_result_pars) * 2

        self._assert_fit_properties(
            _fit,
            dict(
                data=np.array([self._ref_x, self._ref_y_data * 2]),
                x_data=self._ref_x,
                y_data=self._ref_y_data * 2,
                parameter_values=_new_estimates,
            ),
            rtol=1e-2
        )

    def test_update_data_different_length(self):
        _fit = self._get_fit()

        _fit.data = np.array([self._ref_x[:-1], self._ref_y_data[:-1] * 2])
        _fit.add_error('y', err_val=1.0)

        self._assert_fit_properties(
            _fit,
            dict(
                data=np.array([self._ref_x[:-1], self._ref_y_data[:-1] * 2]),
                x_data=self._ref_x[:-1],
                y_data=self._ref_y_data[:-1] * 2,
            )
        )

        _fit.do_fit()

        _new_estimates = analytic_solution(
            self._linear_design_matrix[:-1],
            np.linalg.inv(self._ref_matrix_eye[:-1,:-1]),
            self._ref_y_data[:-1] * 2,
        )

        self._assert_fit_properties(
            _fit,
            dict(
                data=np.array([self._ref_x[:-1], self._ref_y_data[:-1] * 2]),
                x_data=self._ref_x[:-1],
                y_data=self._ref_y_data[:-1] * 2,
                parameter_values=_new_estimates,
            ),
            rtol=1e-2
        )

    def test_reserved_parameter_names_raise(self):
        def dummy_model(x, y_data):
            pass

        with self.assertRaises(ValueError) as _exc:
            XYFit(xy_data=self._ref_xy_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('reserved', _exc.exception.args[0])
        self.assertIn('y_data', _exc.exception.args[0])

    def test_model_no_pars_raise(self):
        def dummy_model():
            pass

        with self.assertRaises(ValueError) as _exc:
            XYFit(xy_data=self._ref_xy_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn(
            'needs at least one parameter',
            _exc.exception.args[0])

    def test_model_no_pars_beside_x_raise(self):
        def dummy_model(x):
            pass

        with self.assertRaises(ValueError) as _exc:
            XYFit(xy_data=self._ref_xy_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn(
            'needs at least one parameter besides',
            _exc.exception.args[0])

    def test_model_varargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            XYFit(xy_data=self._ref_xy_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])

    def test_model_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            XYFit(xy_data=self._ref_xy_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varkwargs', _exc.exception.args[0])

    def test_model_varargs_varkwargs_raise(self):
        # TODO: raise even without 'par'
        def dummy_model(x, par, *varargs, **varkwargs):
            pass

        with self.assertRaises(ValueError) as _exc:
            XYFit(xy_data=self._ref_xy_data,
                  model_function=dummy_model,
                  minimizer=self.MINIMIZER)

        self.assertIn('variable', _exc.exception.args[0])
        self.assertIn('varargs', _exc.exception.args[0])
        # TODO: enable when implemented
        #self.assertIn('varkwargs', _exc.exception.args[0])

    def test_report_before_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_report_after_fit(self):
        # TODO: check report content
        _buffer = six.StringIO()
        _fit = self._get_fit()
        _fit.do_fit()
        _fit.report(output_stream=_buffer)
        self.assertNotEqual(_buffer.getvalue(), "")

    def test_unknown_cost_function(self):
        with self.assertRaises(ValueError):
            self._get_fit(cost_function="ABC123")

    def test_model_as_model_func_name(self):
        def model(x, a, b):
            return a * x + b
        self._get_fit(model_function=model)

    def test_data_and_cost_incompatible(self):
        with self.assertRaises(ValueError):
            self._get_fit(cost_function="nll")

    def test_relative_model_error_nonlinear(self):
        _fit_data_err = self._get_fit(errors=[
            dict(axis="y", err_val=1.0, relative=False, reference="data"),
            dict(axis="y", err_val=0.1, relative=True, reference="data")
        ])
        _fit_model_err = self._get_fit(errors=[
            dict(axis="y", err_val=1.0, relative=False, reference="data"),
            dict(axis="y", err_val=0.1, relative=True, reference="model")
        ], dynamic_error_algorithm="nonlinear")
        _fit_data_err.do_fit()
        _fit_model_err.do_fit()
        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=6e-2)

    def test_relative_model_error_iterative(self):
        _fit_data_err = self._get_fit(errors=[
            dict(axis="y", err_val=1.0, relative=False, reference="data"),
            dict(axis="y", err_val=0.1, relative=True, reference="data")
        ])
        _fit_model_err = self._get_fit(errors=[
            dict(axis="y", err_val=1.0, relative=False, reference="data"),
            dict(axis="y", err_val=0.1, relative=True, reference="model")
        ], dynamic_error_algorithm="iterative")
        _fit_data_err.do_fit()
        _fit_model_err.do_fit()
        self._assert_fit_results_equal(_fit_data_err, _fit_model_err, rtol=1e-2)

    def test_fit_scalar_model_function(self):
        def _model_function(x, a):
            return a

        _fit = XYFit([[0.0, 2.0], [1.0, 2.0]], _model_function)
        _fit.add_error("y", 1.0)
        _fit.do_fit()
        self.assertAlmostEqual(_fit.parameter_values[0], 1.5)
        _fit.report()



class TestXYFitWithSimpleYErrors(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        six.get_unbound_function(TestXYFitBasicInterface.setUp)(self)

    def _get_fit(self, errors=None):
        '''convenience'''

        errors = errors or [dict(axis='y', err_val=1.0)]

        _fit = XYFit(
            xy_data=self._ref_xy_data,
            model_function=simple_xy_model,
            cost_function=XYCostFunction_Chi2(
                axes_to_use='xy', errors_to_use='covariance'),
            minimizer=self.MINIMIZER
        )

        for _err in errors:
            _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            'default': self._get_fit(),
            'two_errors': self._get_fit(
                errors=[
                    dict(axis='y', err_val=1.0/np.sqrt(2)),
                    dict(axis='y', err_val=1.0/np.sqrt(2))
                ]
            ),
            'named_errors': self._get_fit(
                errors=[
                    dict(axis='y', err_val=1.0/np.sqrt(2), name="MyYDataError",
                         correlation=0, relative=False, reference='data'),
                    dict(axis='y', err_val=1.0/np.sqrt(2), name="MyYModelError",
                         correlation=0, relative=False, reference='model'),
                ]
            )
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
            fit_names=['default', 'two_errors'],
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                y_model=self._nominal_fit_result_y_model,
                model=self._nominal_fit_result_xy_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            fit_names=['default', 'two_errors'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_get_matching_error_all(self):
        _fit = self._get_test_fits()['named_errors']
        for _mc in (None, dict()):
            _errs = _fit.get_matching_errors(matching_criteria=_mc)
            self.assertEqual(len(_errs), 2)
            self.assertIs(_fit.data_container._error_dicts['MyYDataError']['err'], _errs['MyYDataError'])
            self.assertIs(_fit._param_model._error_dicts['MyYModelError']['err'], _errs['MyYModelError'])

    def test_get_matching_error_name(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(name='MyYDataError'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts['MyYDataError']['err'], _errs['MyYDataError'])

    def test_get_matching_error_type_simple(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='simple'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MyYDataError']['err'], _errs['MyYDataError'])
        self.assertIs(_fit._param_model._error_dicts['MyYModelError']['err'], _errs['MyYModelError'])

    def test_get_matching_error_type_matrix(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='matrix'))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_uncorrelated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=False))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MyYDataError']['err'], _errs['MyYDataError'])
        self.assertIs(_fit._param_model._error_dicts['MyYModelError']['err'], _errs['MyYModelError'])

    def test_get_matching_error_reference_data(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='data'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts['MyYDataError']['err'], _errs['MyYDataError'])

    def test_get_matching_error_reference_model(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='model'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit._param_model._error_dicts['MyYModelError']['err'], _errs['MyYModelError'])


class TestXYFitWithMatrixErrors(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        six.get_unbound_function(TestXYFitBasicInterface.setUp)(self)

    def _get_fit(self, errors=None):
        '''convenience'''

        _fit = XYFit(
            xy_data=self._ref_xy_data,
            model_function=simple_xy_model,
            cost_function=XYCostFunction_Chi2(
                axes_to_use='xy', errors_to_use='covariance'),
            minimizer=self.MINIMIZER
        )

        if errors is None:
            _fit.add_matrix_error(axis='y', err_matrix=np.eye(self._n_points), matrix_type='cov')
        else:
            for _err in errors:
                if 'err_matrix' in _err:
                    _fit.add_matrix_error(**_err)
                else:
                    _fit.add_error(**_err)

        return _fit

    def _get_test_fits(self):
        return {
            'default': self._get_fit(),
            'cor_matrix_and_error_vector': self._get_fit(
                errors=[
                    dict(axis='y',
                         err_matrix=np.eye(self._n_points),
                         matrix_type='cor',
                         err_val=1.0
                    )
                ]
            ),
            'two_matrix_errors': self._get_fit(
                errors=[
                    dict(axis='y', err_matrix=np.eye(self._n_points)/2, matrix_type='cov'),
                    dict(axis='y', err_matrix=np.eye(self._n_points)/2, matrix_type='cov')
                ]
            ),
            'one_matrix_one_simple_error': self._get_fit(
                errors=[
                    dict(axis='y', err_matrix=np.eye(self._n_points)/2, matrix_type='cov'),
                    dict(axis='y', err_val=1/np.sqrt(2))
                ]
            ),
            'named_errors': self._get_fit(
                errors=[
                    dict(axis='y', err_val=1.0/np.sqrt(2), name="MySimpleDataError",
                         correlation=0, relative=False, reference='data'),
                    dict(axis='y', err_val=1.0/np.sqrt(2), name="MySimpleModelError",
                         correlation=0, relative=False, reference='model'),
                    dict(axis='y', err_matrix=np.eye(self._n_points)/2, matrix_type='cov',
                         name="MyMatrixDataError",
                         relative=False, reference='data'),
                    dict(axis='y', err_matrix=np.eye(self._n_points)/2, matrix_type='cov',
                         name="MyMatrixModelError",
                         relative=False, reference='model'),
                ]
            )
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
            fit_names=['default', 'cor_matrix_and_error_vector',
                       'one_matrix_one_simple_error', 'two_matrix_errors'],
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                y_model=self._nominal_fit_result_y_model,
                model=self._nominal_fit_result_xy_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,
            ),
            fit_names=['default', 'cor_matrix_and_error_vector',
                       'one_matrix_one_simple_error', 'two_matrix_errors'],
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-2
        )

    def test_get_matching_error_all(self):
        _fit = self._get_test_fits()['named_errors']
        for _mc in (None, dict()):
            _errs = _fit.get_matching_errors(matching_criteria=_mc)
            self.assertEqual(len(_errs), 4)
            self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
            self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])
            self.assertIs(_fit.data_container._error_dicts['MyMatrixDataError']['err'], _errs['MyMatrixDataError'])
            self.assertIs(_fit._param_model._error_dicts['MyMatrixModelError']['err'], _errs['MyMatrixModelError'])

    def test_get_matching_error_name(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(name='MySimpleDataError'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])

    def test_get_matching_error_type_simple(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='simple'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
        self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])

    def test_get_matching_error_type_matrix(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(type='matrix'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MyMatrixDataError']['err'], _errs['MyMatrixDataError'])
        self.assertIs(_fit._param_model._error_dicts['MyMatrixModelError']['err'], _errs['MyMatrixModelError'])

    def test_get_matching_error_uncorrelated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(correlated=False))
        # NOTE: passing 'correlated' only matches 'matrix' errors, irrespective of 'True'/'False' value passed
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
        self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])

    def test_get_matching_error_reference_data(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='data'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit.data_container._error_dicts['MySimpleDataError']['err'], _errs['MySimpleDataError'])
        self.assertIs(_fit.data_container._error_dicts['MyMatrixDataError']['err'], _errs['MyMatrixDataError'])

    def test_get_matching_error_reference_model(self):
        _fit = self._get_test_fits()['named_errors']
        _errs = _fit.get_matching_errors(matching_criteria=dict(reference='model'))
        self.assertEqual(len(_errs), 2)
        self.assertIs(_fit._param_model._error_dicts['MySimpleModelError']['err'], _errs['MySimpleModelError'])
        self.assertIs(_fit._param_model._error_dicts['MyMatrixModelError']['err'], _errs['MyMatrixModelError'])


class TestXYFitWithXYErrors(AbstractTestFit, unittest.TestCase):

    MINIMIZER = 'scipy'

    def setUp(self):
        """
        This was changed because the results were suspicious when cov mat determinant was added
        to cost.
        """
        self._n_points = 16
        self._x_error = 0.1
        self._y_error = 0.1
        self._default_cost_function = XYCostFunction_Chi2(
            axes_to_use='xy', errors_to_use='covariance')

        # "jitter" for data smearing
        np.random.seed(0)
        self._x_jitter = np.random.normal(scale=self._x_error, size=self._n_points)
        self._y_jitter = np.random.normal(scale=self._y_error, size=self._n_points)

        assert len(self._y_jitter) == self._n_points

        # reference initial values
        self._ref_initial_pars = np.array([1.0, 0.5])
        self._ref_x_data = np.arange(self._n_points) + self._x_jitter
        self._ref_initial_y_model = line_xy_model(self._ref_x_data, *self._ref_initial_pars)
        self._ref_initial_xy_model = np.array([self._ref_x_data, self._ref_initial_y_model])

        # reference matrices/errors
        self._ref_y_error_vector = np.ones(self._n_points) * self._y_error
        self._ref_y_error_matrix = np.diag(self._ref_y_error_vector ** 2)
        self._ref_x_error_vector = np.ones(self._n_points) * self._x_error
        self._ref_x_error_matrix = np.diag(self._ref_x_error_vector ** 2)

        def line_xy_model_derivative(x, a=1.0, b=0.0):
            return a

        _fp = line_xy_model_derivative(self._ref_x_data, *self._ref_initial_pars)
        self._ref_projected_xy_matrix = (
            self._ref_y_error_matrix +
            self._ref_x_error_matrix * np.outer(_fp, _fp)
        )
        self._ref_projected_xy_errors = np.diag(np.sqrt(self._ref_projected_xy_matrix))
        self._ref_projected_xy_log_determinant = np.log(np.linalg.det(
            self._ref_projected_xy_matrix))

        # fit data
        self._ref_y_data = line_xy_model(self._ref_x_data, *self._ref_initial_pars) + self._y_jitter
        self._ref_xy_data = np.array([self._ref_x_data, self._ref_y_data])

        # pre-fit cost value
        self._ref_initial_cost = self._default_cost_function(
            self._ref_y_data,
            self._ref_initial_y_model,
            np.linalg.cholesky(self._ref_projected_xy_matrix),
            self._ref_initial_pars,
            [],
            self._ref_projected_xy_log_determinant
        )

        # reference fit result values
        self._nominal_fit_result_pars = np.array([1.00441642, 0.48645824])

        self._nominal_fit_result_y_model = line_xy_model(
            self._ref_x_data, *self._nominal_fit_result_pars)
        self._nominal_fit_result_xy_model = np.array(
            [self._ref_x_data, self._nominal_fit_result_y_model])

        self._nominal_fit_result_projected_xy_matrix = self._ref_y_error_matrix\
            + self._nominal_fit_result_pars[0] * self._ref_x_error_matrix
        self._nominal_fit_result_projected_xy_errors = np.diag(np.sqrt(
            self._nominal_fit_result_projected_xy_matrix))
        self._nominal_fit_result_projected_xy_log_determinant = np.log(np.linalg.det(
            self._nominal_fit_result_projected_xy_matrix))

        self._nominal_fit_result_cost = self._default_cost_function(
            self._ref_y_data,
            self._nominal_fit_result_y_model,
            np.linalg.cholesky(self._nominal_fit_result_projected_xy_matrix),
            self._nominal_fit_result_pars,
            [],
            self._nominal_fit_result_projected_xy_log_determinant
        )

        # helper dict with all reference property values
        self._ref_prop_dict = dict(
            parameter_values=self._ref_initial_pars,

            x_data=self._ref_x_data,
            x_model=self._ref_x_data,
            y_data=self._ref_y_data,
            y_model=self._ref_initial_y_model,
            data=self._ref_xy_data,
            model=self._ref_initial_xy_model,
            did_fit=False,

            x_data_cov_mat=self._ref_x_error_matrix,
            y_data_cov_mat=self._ref_y_error_matrix,
            x_data_cov_mat_inverse=np.linalg.inv(self._ref_x_error_matrix),
            y_data_cov_mat_inverse=np.linalg.inv(self._ref_y_error_matrix),
            x_data_cor_mat=np.eye(self._n_points),
            y_data_cor_mat=np.eye(self._n_points),
            x_data_error=self._ref_x_error_vector,
            y_data_error=self._ref_y_error_vector,

            cost_function_value=self._ref_initial_cost,

            total_cov_mat=self._ref_projected_xy_matrix,
            total_error=self._ref_projected_xy_errors,
        )

    def _get_fit(self, errors=None, constraints=None, dynamic_error_algorithm=None):
        '''convenience'''

        errors = errors or [dict(axis='y', err_val=1.0)]
        constraints = constraints or []
        dynamic_error_algorithm = dynamic_error_algorithm or "nonlinear"

        _fit = XYFit(
            xy_data=self._ref_xy_data,
            model_function=line_xy_model,
            cost_function=self._default_cost_function,
            minimizer=self.MINIMIZER,
            dynamic_error_algorithm=dynamic_error_algorithm
        )

        for _err in errors:
            _fit.add_error(**_err)
        for _constraint in constraints:
            _fit.add_parameter_constraint(**_constraint)

        _fit.set_all_parameter_values(self._ref_initial_pars)

        return _fit

    def _get_test_fits(self):
        return {
            'nonlinear_with_x_errors': self._get_fit(
                errors=[
                    dict(axis='x', err_val=self._ref_x_error_vector),
                    dict(axis='y', err_val=self._ref_y_error_vector),
                ]
            ),
        }

    def test_initial_state(self):
        self.run_test_for_all_fits(
            self._ref_prop_dict,
        )

    def test_fit_results(self):
        self.run_test_for_all_fits(
            dict(self._ref_prop_dict,
                parameter_values=self._nominal_fit_result_pars,
                y_model=self._nominal_fit_result_y_model,
                model=self._nominal_fit_result_xy_model,
                did_fit=True,
                cost_function_value=self._nominal_fit_result_cost,

                total_cov_mat=self._nominal_fit_result_projected_xy_matrix,
                total_error=self._nominal_fit_result_projected_xy_errors,

            ),
            call_before_fit=lambda f: f.do_fit(),
            rtol=1e-3,
            atol=1e-2,
        )

    def test_limit_parameter_raise(self):
        with self.assertRaises(ValueError):
            self._get_fit().limit_parameter("a")

    def test_limit_parameter_raise_if_not_numeric(self):
        with self.assertRaises(TypeError):
            # old tuple syntax no longer supported
            self._get_fit().limit_parameter("a", (0, 1))
        with self.assertRaises(TypeError):
            # non-numeric types invalid (except None)
            self._get_fit().limit_parameter("a", "invalid", "limits")
        with self.assertRaises(TypeError):
            # string representations of numerics should also fail
            self._get_fit().limit_parameter("a", "0.3", "14")

    def test_iterative_linear_and_relative_model_error(self):
        _constraints = [
            dict(name="a", value=1.0, uncertainty=0.1),
            dict(name="b", value=0.3, uncertainty=0.1)
        ]
        _errors_rel_data = [
            dict(axis="x", err_val=0.1, relative=False, reference="data"),
            dict(axis="y", err_val=1.0, relative=False, reference="data"),
            dict(axis="y", err_val=0.1, relative=True, reference="data")
        ]
        _errors_rel_model = [
            dict(axis="x", err_val=0.1, relative=False, reference="data"),
            dict(axis="y", err_val=1.0, relative=False, reference="data"),
            dict(axis="y", err_val=0.1, relative=True, reference="model")
        ]
        _fit_rel_data_nonlinear = self._get_fit(
            errors=_errors_rel_data, constraints=_constraints,
            dynamic_error_algorithm="nonlinear")
        _fit_rel_data_iterative = self._get_fit(
            errors=_errors_rel_data, constraints=_constraints,
            dynamic_error_algorithm="iterative")
        _fit_rel_model_nonlinear = self._get_fit(
            errors=_errors_rel_model, constraints=_constraints,
            dynamic_error_algorithm="nonlinear")
        _fit_rel_model_iterative = self._get_fit(
            errors=_errors_rel_model, constraints=_constraints,
            dynamic_error_algorithm="iterative")
        _fit_rel_data_nonlinear.do_fit()
        _fit_rel_data_iterative.do_fit()
        _fit_rel_model_nonlinear.do_fit()
        _fit_rel_model_iterative.do_fit()

        print("============ rel data nonlinear vs rel data iterative ==========\n")
        self._assert_fit_results_equal(
            _fit_rel_data_nonlinear, _fit_rel_data_iterative, rtol=1.25e-2, atol=5e-5)
        print("============ rel data nonlinear vs rel model nonlinear ==========\n")
        self._assert_fit_results_equal(
            _fit_rel_data_nonlinear, _fit_rel_model_nonlinear, rtol=1.25e-2, atol=5e-5)
        print("============ rel data nonlinear vs rel model iterative ==========\n")
        self._assert_fit_results_equal(
            _fit_rel_data_nonlinear, _fit_rel_model_iterative, rtol=1.25e-2, atol=5e-5)
        print("============ rel model nonlinear vs rel model iterative ==========\n")
        self._assert_fit_results_equal(
            _fit_rel_model_nonlinear, _fit_rel_model_iterative, rtol=1.25e-2, atol=5e-5)
        print("============ rel data iterative vs rel model iterative ==========\n")
        self._assert_fit_results_equal(
            _fit_rel_data_iterative, _fit_rel_model_iterative, rtol=1.25e-2, atol=5e-5)
        print("============ rel data iterative vs rel model nonlinear ==========\n")
        self._assert_fit_results_equal(
            _fit_rel_data_iterative, _fit_rel_model_nonlinear, rtol=1.25e-2, atol=5e-5)




import unittest
import numpy as np

from kafe2.fit import IndexedContainer, IndexedParametricModel
from kafe2.fit.indexed.model import IndexedModelFunction
from kafe2.core.error import cov_mat_from_float_list


class TestDatastoreIndexed(unittest.TestCase):

    def setUp(self):
        self._ref_data = [3.3, 5.5, 2.2, 8.5, 10.2]
        self.idx_cont = IndexedContainer(data=self._ref_data)

        self._ref_err_abs_singlevalue = 1.2
        self._ref_err_abs_valuearray = [1.2, 1.1, 1.2, 1.2, 1.1]

        self._ref_err_corr_coeff = 0.23

        self.idx_cont.add_error(self._ref_err_abs_valuearray, name='MyError', correlation=self._ref_err_corr_coeff,
                                relative=False)

        self._ref_cov_mat = cov_mat_from_float_list(self._ref_err_abs_valuearray,
                                                    correlation=self._ref_err_corr_coeff).mat

    def test_get_matching_error_all_empty_dict(self):
        _errs = self.idx_cont.get_matching_errors(matching_criteria=dict())
        self.assertEqual(len(_errs), 1)
        self.assertIs(self.idx_cont._error_dicts['MyError']['err'], _errs['MyError'])

    def test_get_matching_error_all_None(self):
        _errs = self.idx_cont.get_matching_errors(matching_criteria=None)
        self.assertEqual(len(_errs), 1)
        self.assertIs(self.idx_cont._error_dicts['MyError']['err'], _errs['MyError'])

    def test_get_matching_error_name(self):
        _errs = self.idx_cont.get_matching_errors(matching_criteria=dict(name='MyError'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(self.idx_cont._error_dicts['MyError']['err'], _errs['MyError'])

    def test_get_matching_error_type_simple(self):
        _errs = self.idx_cont.get_matching_errors(matching_criteria=dict(type='simple'))
        self.assertEqual(len(_errs), 1)
        self.assertIs(self.idx_cont._error_dicts['MyError']['err'], _errs['MyError'])

    def test_get_matching_error_type_matrix(self):
        _errs = self.idx_cont.get_matching_errors(matching_criteria=dict(type='matrix'))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_uncorrelated(self):
        _errs = self.idx_cont.get_matching_errors(matching_criteria=dict(correlated=False))
        self.assertEqual(len(_errs), 0)

    def test_get_matching_error_correlated(self):
        _errs = self.idx_cont.get_matching_errors(matching_criteria=dict(correlated=True))
        self.assertEqual(len(_errs), 1)
        self.assertIs(self.idx_cont._error_dicts['MyError']['err'], _errs['MyError'])

    def test_compare_error_reference(self):
        for _err_dict in self.idx_cont._error_dicts.values():
            _err_ref_vals = _err_dict['err'].reference
            self.assertTrue(
                np.all(_err_ref_vals == self._ref_data)
            )

    def test_compare_ref_data(self):
        self.assertTrue(np.all(self.idx_cont.data == self._ref_data))

    def test_compare_ref_err(self):
        self.assertTrue(np.allclose(self.idx_cont.err, self._ref_err_abs_valuearray, atol=1e-10))

    def test_compare_ref_total_cov_mat(self):
        _err = self.idx_cont.get_total_error()
        _mat = _err.cov_mat
        self.assertTrue(np.allclose(_mat, self._ref_cov_mat, atol=1e-5))

    def test_compare_ref_total_err_for_add_err_twice(self):
        self.idx_cont.add_error(self._ref_err_abs_valuearray,
                                name='MyNewError',
                                correlation=self._ref_err_corr_coeff,
                                relative=False)
        _err = self.idx_cont.get_total_error()
        _mat = _err.cov_mat
        self.assertTrue(np.allclose(_mat, self._ref_cov_mat + self._ref_cov_mat, atol=1e-5))

    def test_compare_ref_total_err_for_add_disable_new_err(self):
        self.idx_cont.add_error(self._ref_err_abs_valuearray,
                                name='MyNewError',
                                correlation=self._ref_err_corr_coeff,
                                relative=False)
        self.idx_cont.disable_error("MyNewError")
        _err = self.idx_cont.get_total_error()
        _mat = _err.cov_mat

        self.assertTrue(np.allclose(_mat, self._ref_cov_mat, atol=1e-5))

    def test_compare_ref_total_err_for_add_disable_reenable_new_err(self):
        self.idx_cont.add_error(self._ref_err_abs_valuearray,
                                name='MyNewError',
                                correlation=self._ref_err_corr_coeff,
                                relative=False)
        self.idx_cont.disable_error("MyNewError")
        _err = self.idx_cont.get_total_error()
        self.idx_cont.enable_error("MyNewError")
        _err = self.idx_cont.get_total_error()
        _mat = _err.cov_mat

        self.assertTrue(np.allclose(_mat, self._ref_cov_mat + self._ref_cov_mat, atol=1e-5))

    def test_raise_add_same_error_name_twice(self):
        self.idx_cont.add_error(0.1,
                                name="MyNewError",
                                correlation=0, relative=False)
        with self.assertRaises(ValueError):
            self.idx_cont.add_error(0.1,
                                    name="MyNewError",
                                    correlation=0, relative=False)

    def test_raise_get_inexistent_error(self):
        with self.assertRaises(ValueError):
            self.idx_cont.get_error("MyInexistentError")

    def test_raise_add_error_bad_shape(self):
        with self.assertRaises(ValueError):
            self.idx_cont.add_error([1, 2, 3])

    def test_raise_add_matrix_error_bad_shape(self):
        with self.assertRaises(ValueError):
            self.idx_cont.add_matrix_error([[1, 0], [0, 4]], "cov")
        with self.assertRaises(ValueError):
            self.idx_cont.add_matrix_error(np.eye(5), "cor", err_val=np.ones(10))
        with self.assertRaises(ValueError):
            self.idx_cont.add_matrix_error(np.eye(5), "cor", err_val=np.eye(5))

    def test_raise_bad_data_shape(self):
        with self.assertRaises(ValueError):
            _ = IndexedContainer([[1, 2], [3, 4]])

class TestDatastoreIndexedParametricModel(unittest.TestCase):
    def _ref_model_func(self, slope, intercept):
        return slope * self._ref_pm_support + intercept

    def _ref_model_func_deriv_by_pars(self, slope, intercept):
        return np.array([self._ref_pm_support, [1]*len(self._ref_pm_support)])

    def setUp(self):
        self._ref_pm_support = np.linspace(-5, 5, 11)

        self._ref_params = (1.2, 3.3)
        self._ref_data = self._ref_model_func(*self._ref_params)

        self.idx_param_model = IndexedParametricModel(
            model_func=IndexedModelFunction(self._ref_model_func), 
            model_parameters=self._ref_params)

        self._test_params = (3.4, -5.23)
        self._ref_test_data = self._ref_model_func(*self._test_params)

    def test_compare_ref_data(self):
        self.assertTrue(np.all(self.idx_param_model.data == self._ref_data))

    def test_deriv_by_par(self):
        self.assertTrue(
            np.allclose(
                self.idx_param_model.eval_model_function_derivative_by_parameters(),
                self._ref_model_func_deriv_by_pars(*self._ref_params)
            )
        )


    def test_change_parameters_test_data(self):
        self.idx_param_model.parameters = self._test_params
        self.assertTrue(np.all(self.idx_param_model.data == self._ref_test_data))

    def test_raise_set_data(self):
        with self.assertRaises(TypeError):
            self.idx_param_model.data = self._ref_test_data



"""This submodule provides the necessary objects for parameter estimation from unbinned datasets. Those fits are used
when there are two few data points to fill a histogram. Unbinned fits are essentially the ground truth for histogram
fits as there is no modification of the data points through the binning. As each data point is considered, performing an
unbinned fit can take much longer than a histogram fit, where each bin is considered.

:synopsis: This submodule provides the necessary objects for parameter estimation from unbinned datasets.

.. moduleauthor:: Cedric Verstege <cedric.verstege@student.kit.edu>
"""

from .container import *
from .cost import *
from .fit import *
from .model import *
from .plot import *



import numpy as np

from scipy.misc import derivative

from .._base import ParametricModelBaseMixin, ModelFunctionBase
from .container import IndexedContainer
from .format import IndexedModelFunctionFormatter


__all__ = ["IndexedParametricModel", "IndexedModelFunction"]


class IndexedModelFunction(ModelFunctionBase):
    FORMATTER_TYPE = IndexedModelFunctionFormatter

    def __init__(self, model_function):
        """
        Construct :py:class:`IndexedModelFunction` object (a wrapper for a native Python function):

        :param model_function: function handle
        """
        # Indexed Fit Functions don't have independent arguments. They solely consist of parameters to be fitted.
        super(IndexedModelFunction, self).__init__(model_function=model_function, independent_argcount=0)


class IndexedParametricModel(ParametricModelBaseMixin, IndexedContainer):
    MODEL_FUNCTION_TYPE = IndexedModelFunction

    def __init__(self, model_func, model_parameters, shape_like=None):
        """
        Construct an :py:obj:`IndexedParametricModel` object:

        :param model_func: handle of Python function (the model function)
        :param model_parameters: iterable of parameter values with which the model function should be initialized
        :param shape_like: array with the same shape as the model
        """
        # print "IndexedParametricModel.__init__(model_func=%r, model_parameters=%r)" % (model_func, model_parameters)
        if shape_like is not None:
            _data = np.zeros_like(shape_like)
            try:
                _data[:] = model_func(*model_parameters)
            except ValueError as _e:
                raise ValueError("Indexed Data and Function must have the same shape! Got {} and {}"
                                 .format(len(_data), len(model_func(*model_parameters)))) from _e
        else:
            _data = model_func(*model_parameters)
        super(IndexedParametricModel, self).__init__(model_func, model_parameters, _data)

    # -- private methods

    def _recalculate(self):
        # use parent class setter for 'data'
        IndexedContainer.data.fset(self, self.eval_model_function())
        self._pm_calculation_stale = False


    # -- public properties

    @property
    def data(self):
        """model predictions (one-dimensional :py:obj:`numpy.ndarray`)"""
        if self._pm_calculation_stale:
            self._recalculate()
        return super(IndexedParametricModel, self).data

    @data.setter
    def data(self, new_data):
        raise TypeError("Parametric model data cannot be set!")

    @property
    def data_range(self):
        """tuple containing the minimum and maximum of all model predictions"""
        _data = self.data
        return np.min(_data), np.max(_data)

    # -- public methods

    def eval_model_function(self, model_parameters=None):
        """
        Evaluate the model function.

        :param model_parameters: values of the model parameters (if ``None``, the current values are used)
        :type model_parameters: list or ``None``
        :return: value(s) of the model function for the given parameters
        :rtype: :py:obj:`numpy.ndarray`
        """
        _pars = model_parameters if model_parameters is not None else self._model_parameters
        return self._model_function_object(*_pars)

    def eval_model_function_derivative_by_parameters(self, model_parameters=None, par_dx=None):
        """
        Evaluate the derivative of the model function with respect to the model parameters.

        :param model_parameters: values of the model parameters (if ``None``, the current values are used)
        :type model_parameters: list or ``None``
        :param par_dx: step size for numeric differentiation
        :type par_dx: float
        :return: value(s) of the model function derivative for the given parameters
        :rtype: :py:obj:`numpy.ndarray`
        """
        _pars = model_parameters if model_parameters is not None else self._model_parameters
        _pars = np.asarray(_pars)
        _par_dxs = par_dx if par_dx is not None else 1e-2 * (np.abs(_pars) + 1.0 / (1.0 + np.abs(_pars)))

        _ret = []
        for _par_idx, (_par_val, _par_dx) in enumerate(zip(_pars, _par_dxs)):
            def _chipped_func(par):
                _chipped_pars = _pars.copy()
                _chipped_pars[_par_idx] = par
                return self._model_function_object(*_chipped_pars)

            _der_val = derivative(_chipped_func, _par_val, dx=_par_dx)
            _ret.append(_der_val)
        return np.array(_ret)



import unittest
import numpy as np

from kafe2.core.constraint import GaussianMatrixParameterConstraint, GaussianSimpleParameterConstraint


class TestMatrixParameterConstraintDirect(unittest.TestCase):

    def setUp(self):
        self._fit_par_values = [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.0]
        self._par_test_values = np.array([[8.9, 3.4, 5.6], [0.1, 2.3, 9.0], [5.6, 4.5, 3.4]])
        self._par_indices = [[8, 3, 5], [0, 2, 9], [5, 4, 3]]
        self._par_values = np.array([[1.23, 7.20, 3.95], [4.11, 3.00, 2.95], [0.1, -8.5, 67.0]])
        self._par_cov_mats_abs = np.array([
            [
                [1.0, 0.0, 0.0],
                [0.0, 2.8, 0.0],
                [0.0, 0.0, 0.5],
            ], [
                [1.0, 0.2, 0.3],
                [0.2, 2.8, 0.1],
                [0.3, 0.1, 0.5],
            ]
        ])
        self._par_cov_mats_rel = np.array([
            [
                [0.1, 0.0, 0.0],
                [0.0, 0.2, 0.0],
                [0.0, 0.0, 0.3],
            ], [
                [0.10, 0.01, 0.02],
                [0.01, 0.20, 0.03],
                [0.02, 0.03, 0.30],
            ]
        ])
        self._uncertainties_abs = np.array([[1.2, 2.3, 0.4], [6.5, 2.6, 1.0]])
        self._uncertainties_rel = np.array([[0.2, 0.3, 0.2], [0.5, 2.6, 1.0]])
        self._par_cor_mats = np.array([
            [
                [1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0]
            ], [
                [1.0, 0.1, 0.2],
                [0.1, 1.0, 0.3],
                [0.2, 0.3, 1.0]
            ]
        ])

        self._expected_cost_cov_abs = np.zeros((3, 3, 2))
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    _res = self._par_test_values[_i] - self._par_values[_j]
                    self._expected_cost_cov_abs[_i, _j, _k] = _res.dot(np.linalg.inv(self._par_cov_mats_abs[_k])).dot(_res)
        self._expected_cost_cov_rel = np.zeros((3, 3, 2))
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    _res = self._par_test_values[_i] - self._par_values[_j]
                    _abs_cov_mat = self._par_cov_mats_rel[_k] * np.outer(self._par_values[_j], self._par_values[_j])
                    self._expected_cost_cov_rel[_i, _j, _k] = _res.dot(np.linalg.inv(_abs_cov_mat)).dot(_res)
        self._expected_cost_cor_abs = np.zeros((3, 3, 2, 2))
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    for _l in range(2):
                        _res = self._par_test_values[_i] - self._par_values[_j]
                        _cov_mat = self._par_cor_mats[_k] * np.outer(self._uncertainties_abs[_l],
                                                                         self._uncertainties_abs[_l])
                        self._expected_cost_cor_abs[_i, _j, _k, _l] = _res.dot(np.linalg.inv(_cov_mat)).dot(_res)
        self._expected_cost_cor_rel = np.zeros((3, 3, 2, 2))
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    for _l in range(2):
                        _res = self._par_test_values[_i] - self._par_values[_j]
                        _uncertainties_abs = self._uncertainties_rel[_l] * self._par_values[_j]
                        _cov_mat = self._par_cor_mats[_k] * np.outer(_uncertainties_abs, _uncertainties_abs)
                        self._expected_cost_cor_rel[_i, _j, _k, _l] = _res.dot(np.linalg.inv(_cov_mat)).dot(_res)

    def _call_all_properties(self, matrix_constraint):
        matrix_constraint.indices
        matrix_constraint.values
        matrix_constraint.cov_mat
        matrix_constraint.cov_mat_rel
        matrix_constraint.cor_mat
        matrix_constraint.uncertainties
        matrix_constraint.uncertainties_rel
        matrix_constraint.matrix_type
        matrix_constraint.relative
        matrix_constraint.cov_mat_inverse

    def test_bad_input_errors(self):
        with self.assertRaises(ValueError):  # matrix not symmetric
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0], matrix=[[0.5, 0.1], [0.0, 0.6]])
        with self.assertRaises(ValueError):  # values wrong dim
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[[0.0, 1.0]], matrix=[[0.5, 0.0], [0.0, 0.6]])
        with self.assertRaises(ValueError):  # values wrong length
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0, 5.0], matrix=[[0.5, 0.0], [0.0, 0.6]])
        with self.assertRaises(ValueError):  # both uncertainties and cov mat
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0], matrix=[[0.5, 0.0], [0.0, 0.6]],
                                              uncertainties=[0.5, 0.6])
        with self.assertRaises(ValueError):  # unknown matrix type
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0], matrix=[[0.5, 0.0], [0.0, 0.6]],
                                              matrix_type='cost')
        with self.assertRaises(ValueError):  # cor_mat but no uncertainties
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0], matrix=[[1.0, 0.0], [0.0, 1.0]],
                                              matrix_type='cor')
        with self.assertRaises(ValueError):  # cor mat diagonal elements != 1
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0], matrix=[[1.1, 0.0], [0.0, 1.0]],
                                              matrix_type='cor')
        with self.assertRaises(ValueError):  # cor mat elements > 1
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0], matrix=[[1.0, 2.0], [2.0, 1.0]],
                                              matrix_type='cor')
        with self.assertRaises(ValueError):  # cor mat elements < -1
            GaussianMatrixParameterConstraint(indices=[1, 0], values=[0.0, 1.0], matrix=[[1.0, -2.0], [-2.0, 1.0]],
                                              matrix_type='cor')

    def test_cost_matrix_cov_abs(self):
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    _constraint = GaussianMatrixParameterConstraint(
                        self._par_indices[_i], self._par_values[_j], self._par_cov_mats_abs[_k])
                    self.assertTrue(np.abs(
                        _constraint.cost(self._fit_par_values) - self._expected_cost_cov_abs[_i, _j, _k]) < 1e-12)
                    self._call_all_properties(_constraint)

    def test_cost_matrix_cov_rel(self):
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    _constraint = GaussianMatrixParameterConstraint(
                        self._par_indices[_i], self._par_values[_j], self._par_cov_mats_rel[_k],
                        relative=True
                    )
                    self.assertTrue(np.abs(
                        _constraint.cost(self._fit_par_values) - self._expected_cost_cov_rel[_i, _j, _k]) < 1e-12)
                    self._call_all_properties(_constraint)

    def test_cost_matrix_cor_abs(self):
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    for _l in range(2):
                        _constraint = GaussianMatrixParameterConstraint(
                            self._par_indices[_i], self._par_values[_j], self._par_cor_mats[_k],
                            matrix_type='cor', uncertainties=self._uncertainties_abs[_l]
                        )
                        self.assertTrue(np.abs(
                            _constraint.cost(self._fit_par_values)
                            - self._expected_cost_cor_abs[_i, _j, _k, _l]) < 1e-12)
                        self._call_all_properties(_constraint)

    def test_cost_matrix_cor_rel(self):
        for _i in range(3):
            for _j in range(3):
                for _k in range(2):
                    for _l in range(2):
                        _constraint = GaussianMatrixParameterConstraint(
                            self._par_indices[_i], self._par_values[_j], self._par_cor_mats[_k],
                            matrix_type='cor', uncertainties=self._uncertainties_rel[_l], relative=True
                        )
                        self.assertTrue(np.abs(
                            _constraint.cost(self._fit_par_values)
                            - self._expected_cost_cor_rel[_i, _j, _k, _l]) < 1e-12)
                        self._call_all_properties(_constraint)


class TestSimpleParameterConstraintDirect(unittest.TestCase):

    def setUp(self):
        self._fit_par_values = [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.0]
        self._par_test_values = [8.9, 3.4, 5.6]
        self._par_indices = [8, 3, 5]
        self._par_values = [1.23, 7.20, 3.95]
        self._par_uncertainties_abs = [1.0, 2.8, 0.001]
        self._par_uncertainties_rel = [0.1, 0.3, 0.01]
        self._expected_cost_abs = np.zeros((3, 3, 3))
        for _i in range(3):
            for _j in range(3):
                for _k in range(3):
                    _res = self._par_test_values[_i] - self._par_values[_j]
                    self._expected_cost_abs[_i, _j, _k] = (_res / self._par_uncertainties_abs[_k]) ** 2
        self._expected_cost_rel = np.zeros((3, 3, 3))
        for _i in range(3):
            for _j in range(3):
                for _k in range(3):
                    _res = self._par_test_values[_i] - self._par_values[_j]
                    self._expected_cost_rel[_i, _j, _k] = (_res / (self._par_uncertainties_rel[_k]
                                                                   * self._par_values[_j])) ** 2

    def test_cost_simple_abs(self):
        for _i in range(3):
            for _j in range(3):
                for _k in range(3):
                    _constraint = GaussianSimpleParameterConstraint(
                        self._par_indices[_i], self._par_values[_j], self._par_uncertainties_abs[_k])
                    self.assertTrue(np.allclose(
                        _constraint.cost(self._fit_par_values), self._expected_cost_abs[_i, _j, _k]))

                    # ensure that results are consistent with matrix constraints
                    _constraint = GaussianMatrixParameterConstraint(
                        [self._par_indices[_i]], [self._par_values[_j]], [[self._par_uncertainties_abs[_k] ** 2]]
                    )
                    self.assertTrue(np.allclose(
                        _constraint.cost(self._fit_par_values), self._expected_cost_abs[_i, _j, _k]))
                    _constraint = GaussianMatrixParameterConstraint(
                        [self._par_indices[_i]], [self._par_values[_j]], [[1.0]],
                        matrix_type='cor', uncertainties=[self._par_uncertainties_abs[_k]]
                    )
                    self.assertTrue(np.allclose(
                        _constraint.cost(self._fit_par_values), self._expected_cost_abs[_i, _j, _k]))

    def test_cost_simple_rel(self):
        for _i in range(3):
            for _j in range(3):
                for _k in range(3):
                    _constraint = GaussianSimpleParameterConstraint(
                        self._par_indices[_i], self._par_values[_j], self._par_uncertainties_rel[_k],
                        relative=True
                    )
                    self.assertTrue(np.allclose(
                        _constraint.cost(self._fit_par_values), self._expected_cost_rel[_i, _j, _k]))

                    # ensure that results are consistent with matrix constraints
                    _constraint = GaussianMatrixParameterConstraint(
                        [self._par_indices[_i]], [self._par_values[_j]], [[self._par_uncertainties_rel[_k] ** 2]],
                        relative=True
                    )
                    self.assertTrue(np.allclose(
                        _constraint.cost(self._fit_par_values), self._expected_cost_rel[_i, _j, _k]))
                    _constraint = GaussianMatrixParameterConstraint(
                        [self._par_indices[_i]], [self._par_values[_j]], [[1.0]],
                        matrix_type='cor', uncertainties=[self._par_uncertainties_rel[_k]], relative=True
                    )
                    self.assertTrue(np.allclose(
                        _constraint.cost(self._fit_par_values), self._expected_cost_rel[_i, _j, _k]))



import numpy as np
#TODO documentation

__all__ = ['linear_model', 'linear_model_derivative', 'quadratic_model', 'quadratic_model_derivative',
           'cubic_model', 'cubic_model_derivative', 'exponential_model', 'normal_distribution']


def linear_model(x, a, b):
    return a * x + b


def linear_model_derivative(x, a, b):
    return np.ones_like(x) * a


linear_model.expression_format_string = "{a} * {x} + {b}"
linear_model.latex_expression_format_string = r"{a} \, {x} + {b}"


def quadratic_model(x, a, b, c):
    return a * x ** 2 + b * x + c


def quadratic_model_derivative(x, a, b, c):
    return 2 * a * x + b


quadratic_model.expression_format_string = "{a} * {x} ** 2 + {b} * {x} + {c}"
quadratic_model.latex_expression_format_string = r"{a} \, {x}^2 + {b} \, {x} + {c}"


def cubic_model(x, a, b, c, d):
    return a * x ** 3 + b * x ** 2 + c * x + d


def cubic_model_derivative(x, a, b, c, d):
    return 3 * a * x ** 2 + 2 * b * x + c


cubic_model.expression_format_string = "{a} * {x} ** 3 + {b} * {x} ** 2 + {c} * {x} + {d}"
cubic_model.latex_expression_format_string = r"{a} \, {x}^3 + {b} \, {x}^2 + {c} \, {x} + {d}"


def exponential_model(x, A_0, x_0):
    return A_0 * np.exp(x / x_0)


def exponential_model_derivative(x, A_0, x_0):
    return A_0 * np.exp(x / x_0)


exponential_model.expression_format_string = "{A_0} * exp({x} / {x_0})"
exponential_model.latex_expression_format_string = r"{A_0} \, e^{{{x} \ / {x_0}}}"


def normal_distribution(x, mu, sigma):
    return np.exp(-0.5 * ((x - mu) / sigma) ** 2) / np.sqrt(2.0 * np.pi * sigma ** 2)


normal_distribution.expression_format_string = \
    "1 / sqrt(2 * pi * {sigma}) * exp(-0.5 * (({x} - {mu}) / {sigma}) ** 2)"
normal_distribution.latex_expression_format_string = \
    (r"\frac{{1}}{{2 \pi {sigma}}} e^{{- \frac{{1}}{{2}}"
    r"\left( \frac{{{x} - {mu}}}{{{sigma}}} \right)^2 }}")

normal_distribution_pdf = normal_distribution  # Backwards compatibility


STRING_TO_FUNCTION = {
    'line': linear_model,
    'linear': linear_model,
    'linear_model': linear_model,
    'quadratic': quadratic_model,
    'quadratic_model': quadratic_model,
    'cubic': cubic_model,
    'cubic_model': cubic_model,
    'exp': exponential_model,
    'exponential': exponential_model,
    'exponential_model': exponential_model,
    'normal': normal_distribution,
    'normal_distribution': normal_distribution,
    'normal_distribution_pdf': normal_distribution,
}



import numpy as np
import six
from scipy import integrate
from inspect import getsource

from .._base import ParametricModelBaseMixin, ModelFunctionBase
from .container import HistContainer
from ..util import function_library


if six.PY2:
    from funcsigs import signature
else:
    from inspect import signature


__all__ = ["HistParametricModel", "HistModelFunction"]


class HistModelFunction(ModelFunctionBase):

    def __init__(self, model_function=None):
        """
        Construct :py:class:`XYModelFunction` object (a wrapper for a native Python function):

        :param model_function: function handle
        """
        # TODO: default model function
        super(HistModelFunction, self).__init__(
            model_function=model_function, independent_argcount=1)


class HistParametricModel(ParametricModelBaseMixin, HistContainer):
    MODEL_FUNCTION_TYPE = HistModelFunction

    #TODO n_bins, bin_range, bin_edges contain redundant information, should the arguments for HistParametricModel be refactored?
    def __init__(self, n_bins, bin_range,
                 model_density_func=function_library.normal_distribution,
                 model_parameters=[1.0, 1.0], bin_edges=None, bin_evaluation="simpson",
                 density=True):
        super(HistParametricModel, self).__init__(
            model_density_func, model_parameters, n_bins, bin_range,
            bin_edges=bin_edges, fill_data=None, dtype=float)
        self._bin_evaluation = bin_evaluation
        if isinstance(self._bin_evaluation, str):
            self._bin_evaluation = self._bin_evaluation.lower()
            self._bin_evaluation_string = self._bin_evaluation
            if self._bin_evaluation in ("rectangle", "midpoint"):
                self._bin_evaluation_method = self._bin_evaluation_rectangle
            elif self._bin_evaluation == "trapezoid":
                self._bin_evaluation_method = self._bin_evaluation_trapezoid
            elif self._bin_evaluation == "simpson":
                self._bin_evaluation_method = self._bin_evaluation_simpson
            elif self._bin_evaluation == "numerical":
                self._bin_evaluation_method = self._bin_evaluation_numerical
            else:
                raise ValueError("Unknown bin evaluation method: %s" % self._bin_evaluation)
        else:
            if isinstance(self._bin_evaluation, np.vectorize):
                # special handling of numpy vectorized antiderivative functions
                _antiderivative_function_handle = self._bin_evaluation.pyfunc
            elif callable(self._bin_evaluation):
                # handle generic callables
                _antiderivative_function_handle = self._bin_evaluation
            else:
                # raise if not string and not callable
                raise ValueError(
                    "Cannot use {} as bin evaluation method: not a string and not callable!".format(
                        self._bin_evaluation))

            # Retrieving source code will fail if the function was generated through exec.
            # For kafe2go self._bin_evaluation_string will be replaced.
            try:
                self._bin_evaluation_string = getsource(_antiderivative_function_handle)
            except OSError:
                self._bin_evaluation_string = "OSError"
            except IOError:
                self._bin_evaluation_string = "IOError"
            _antider_parameters = list(signature(_antiderivative_function_handle).parameters)
            if isinstance(self._model_function_object, HistModelFunction):
                _model_func_parameters = list(self._model_function_object.signature.parameters)
            else:
                _model_func_parameters = list(signature(self._model_function_object).parameters)

            # require antiderivative and density to have the same arguments
            if _model_func_parameters != _antider_parameters:
                raise ValueError(
                    "Model density function and its antiderivative have different argument "
                    "signatures: (%r vs %r)"
                    % (_model_func_parameters, _antider_parameters))
            self._bin_evaluation_method = self._bin_evaluation_antiderivative

        self._density = density
    # -- private methods

    def _recalculate(self):
        # don't use parent class setter for 'data' -> set directly
        self._data[1:-1] = self._bin_evaluation_method()
        self._pm_calculation_stale = False

    def _bin_evaluation_rectangle(self):
        _height_centers = self.eval_model_function_density(self.bin_centers)
        return self.bin_widths * _height_centers

    def _bin_evaluation_trapezoid(self):
        _height_edges = self.eval_model_function_density(self._bin_edges)
        return self.bin_widths / 2.0 * (_height_edges[:-1] + _height_edges[1:])

    def _bin_evaluation_simpson(self):
        _height_edges = self.eval_model_function_density(self._bin_edges)
        _height_centers = self.eval_model_function_density(self.bin_centers)
        return self.bin_widths / 6.0 * (
                _height_edges[:-1] + 4.0 * _height_centers + _height_edges[1:])

    def _bin_evaluation_numerical(self):
        _integrand_func = lambda x: self.eval_model_function_density(x)
        _int_val = np.zeros(self.size)
        for _i, (_a, _b) in enumerate(zip(self._bin_edges[:-1], self._bin_edges[1:])):
            _int_val[_i], _ = integrate.quad(_integrand_func, _a, _b)
        return _int_val

    def _bin_evaluation_antiderivative(self):
        _fval_antider_as = self._bin_evaluation(self._bin_edges[:-1], *self._model_parameters)
        _fval_antider_bs = self._bin_evaluation(self._bin_edges[1:], *self._model_parameters)
        assert len(_fval_antider_as) == len(_fval_antider_bs) == self.size
        return np.asarray(_fval_antider_bs) - np.asarray(_fval_antider_as)

    # -- public properties

    @property
    def data(self):
        """model predictions (one-dimensional :py:obj:`numpy.ndarray`)"""
        if self._pm_calculation_stale:
            self._recalculate()
        return super(HistParametricModel, self).data

    @data.setter
    def data(self, new_data):
        raise TypeError("Parametric model data cannot be set!")

    @property
    def bin_evaluation(self):
        """
        :return: how the model evaluates bin heights.
        :rtype str, callable, or numpy.vectorize
        """
        return self._bin_evaluation

    @property
    def bin_evaluation_string(self):
        """
        :return: string representation of how the model evaluates bin heights.
        :rtype str
        """
        return self._bin_evaluation_string

    @property
    def density(self):
        return self._density

    # -- public methods

    def eval_model_function_density(self, x, model_parameters=None):
        """
        Evaluate the model function density.

        :param x: *x* values of the support points
        :type x: list of float
        :param model_parameters: values of the model parameters (if ``None``, the current values are used)
        :type model_parameters: list or ``None``
        :return: value(s) of the model function for the given parameters
        :rtype: :py:obj:`numpy.ndarray`
        """
        _pars = model_parameters if model_parameters is not None else self._model_parameters
        _y = self._model_function_object(x, *_pars)
        if np.isscalar(_y):
            _y = np.ones_like(x) * _y
        return _y

    def fill(self, entries):
        raise TypeError("Parametric model of histogram cannot be filled!")



"""This submodule provides the necessary objects for parameter estimation using only a cost function
without explicit data and model. Because of this kafe2 Plots are not available, though plots of the
parameter profiles and contours can still be created.

:synopsis: This submodule provides the necessary objects for parameter estimation using only a cost
function without explicit data and model.

.. moduleauthor:: Johannes Gaessler <johannes.gaessler@student.kit.edu>
"""

from .fit import *



from .._base import FunctionFormatter
from .._base.format import latexify_ascii


__all__ = ["IndexedModelFunctionFormatter"]


class IndexedModelFunctionFormatter(FunctionFormatter):
    def __init__(self, name, latex_name=None, index_name='i', latex_index_name='i',
                 arg_formatters=None, expression_string=None, latex_expression_string=None):
        """
        Construct a :py:obj:`Formatter` for a model function for *indexed* data:

        :param name: a plain-text-formatted string indicating the function name
        :param latex_name: a LaTeX-formatted string indicating the function name
        :param index_name: a plain-text-formatted string representing the index
        :param latex_index_name: a LaTeX-formatted string representing the index
        :param arg_formatters: list of :py:obj:`ParameterFormatter`-derived objects,
                               formatters for function arguments
        :param expression_string:  a plain-text-formatted string indicating the function expression
        :param latex_expression_string:  a LaTeX-formatted string indicating the function expression
        """
        self.index_name = index_name
        self.latex_index_name = latex_index_name
        super(IndexedModelFunctionFormatter, self).__init__(name, latex_name=latex_name, arg_formatters=arg_formatters,
                                                            expression_string=expression_string,
                                                            latex_expression_string=latex_expression_string)

    def _get_format_kwargs(self, format_as_latex=False):
        """Create a dictionary containing argument name and format pairs.

        :param format_as_latex: If the format string is a latex formatted string.
        :return: Dictionary containing argument name and format pairs.
        """
        if format_as_latex:
            _par_name_string_dict = {_af.name: _af.latex_name for _af in self.arg_formatters}
            _par_name_string_dict[self.index_name] = self.latex_index_name
        else:
            _par_name_string_dict = {_af.name: _af.name for _af in self.arg_formatters}
            _par_name_string_dict[self.index_name] = self.index_name
        return _par_name_string_dict

    @property
    def index_name(self):
        """The parameter name of the index.

        :rtype: str
        """
        return self._index_name

    @index_name.setter
    def index_name(self, index_name):
        self._index_name = index_name

    @property
    def latex_index_name(self):
        """The LaTeX parameter name of the index.

        :rtype: str"""
        return self._latex_index_name

    @latex_index_name.setter
    def latex_index_name(self, latex_index_name):
        if latex_index_name is None:
            self._latex_index_name = latexify_ascii(self.index_name)
        elif latex_index_name.startswith('{') and latex_index_name.endswith('}'):
            self._latex_index_name = latex_index_name
        else:
            self._latex_index_name = '{'+latex_index_name+'}'

    def get_formatted(self, with_par_values=False, n_significant_digits=2, format_as_latex=False,
                      with_expression=False):
        """
        Get a formatted string representing this model function.

        :param with_par_values: if ``True``, output will include the value of each function parameter
                                (e.g. ``f_i(a=1, b=2, ...)``)
        :param n_significant_digits: number of significant digits for rounding
        :type n_significant_digits: int
        :param format_as_latex: if ``True``, the returned string will be formatted using LaTeX syntax
        :param with_expression: if ``True``, the returned string will include the expression assigned to the function
        :return: string
        """
        _par_strings = self._get_formatted_pars(with_par_values=with_par_values,
                                                n_significant_digits=n_significant_digits,
                                                format_as_latex=format_as_latex)
        _par_expr_string = ""
        if with_expression:
            _par_expr_string = self._get_formatted_expression(format_as_latex=format_as_latex)

        if format_as_latex:
            _out_string = r"{%s}_{%s}\left(%s\right)" % (self._latex_name, self.latex_index_name, ", ".join(_par_strings))
            if _par_expr_string:
                _out_string += " = " + _par_expr_string
            _out_string = "$%s$" % (_out_string,)
        else:
            _out_string = "%s_%s(%s)" % (self._name, self.index_name, ", ".join(_par_strings))
            if _par_expr_string:
                _out_string += " = " + _par_expr_string
        return _out_string



"""Handle global kafe2 configuration.
"""

import os
import yaml
# WARNING! DO NOT REMOVE THE PYPLOT IMPORT, EVEN THOUGH IT IS NOT USED.
# force pyplot import once to trigger eventual hooks for backend change, e.g. when running inside jupyter
# those hooks are not triggered when only importing matplotlib
import matplotlib.pyplot as plt
import matplotlib as mpl
from copy import deepcopy
from tempfile import NamedTemporaryFile

import logging
logger = logging.getLogger(__name__)

__all__ = ['ConfigLoader', 'ConfigError', 'ConfigTypeError', 'ConfigLookupError', 'kafe2_rc', 'kc']


class ConfigError(Exception):
    pass


class ConfigLookupError(ConfigError):
    def __init__(self, key_path, problematic_key_index):
        self.message = "Error getting config key for '{}': no node under that path!".format(', '.join(key_path))
        self.message = ("Error getting config key for '{}': "
                        "cannot find config node '{}'!".format(', '.join(key_path), key_path[problematic_key_index]))


class ConfigTypeError(ConfigError):
    def __init__(self, key_path, problematic_key_index):
        self.message = ("Error getting config key for '{}': "
                        "scalar node '{}' encountered inside path!".format(', '.join(key_path),
                                                                           key_path[problematic_key_index]))


class ConfigLoader(yaml.Loader):
    def __init__(self, file_like):
        self._current_dir = os.path.split(file_like.name)[0]
        super(ConfigLoader, self).__init__(file_like)

    # custom directives
    def include(self, node):
        _include_path = self.construct_scalar(node)
        with open(os.path.join(self._current_dir, _include_path)) as _f:
            return yaml.load(_f, ConfigLoader)


ConfigLoader.add_constructor('!include', ConfigLoader.include)

# -- read in default global kafe2 configuration from file
with open(os.path.join(__path__[0], 'kafe2.yaml')) as _f:
    _kc = yaml.load(_f, ConfigLoader)

# make a copy of the default configuration
_default_kc = deepcopy(_kc)

# look for local config files in order to overwrite default config
for ext in ('yaml', 'yml'):
    try:
        with open(os.path.join(os.getcwd(), 'kafe2.'+ext)) as _f:
            _kc.update(yaml.load(_f, ConfigLoader))
    except IOError:
        pass


def kc(*keys):
    """Lookup configuration entry by providing the path to it."""
    _dict = _kc

    # if called without args, retrieve the entire configuration dict
    if not keys:
        return _kc

    try:
        for _i, _k in enumerate(keys):
            _dict = _dict[_k]
    except KeyError:
        raise ConfigLookupError(keys, _i)
    except TypeError:
        raise ConfigTypeError(keys, _i)

    return _dict


kafe2_rc = None  # if mpl is only a mock module
try:
    if mpl.__version__.startswith('2'):
        kafe2_rc = mpl.rc_params_from_file(os.path.join(__path__[0], 'kafe2.matplotlibrc.conf'))
        try:  # look for local config
            kafe2_rc.update(mpl.rc_params_from_file(os.path.join(os.getcwd(), 'kafe2.matplotlibrc.conf'),
                                                    use_default_template=False))
        except IOError:
            pass
    elif mpl.__version__.startswith('3'):
        _temp_file = NamedTemporaryFile(delete=False)
        with open(os.path.join(__path__[0], 'kafe2.matplotlibrc.conf')) as _file:
            for _line in _file.readlines():
                if _line.startswith("text.latex.unicode"):
                    continue
                _temp_file.write(_line.encode())
                _temp_file.write('\n'.encode())
        kafe2_rc = mpl.rc_params_from_file(_temp_file.name)
        try:  # look for local config, assume correct mpl3 handling
            kafe2_rc.update(mpl.rc_params_from_file(os.path.join(os.getcwd(), 'kafe2.matplotlibrc.conf'),
                                                    use_default_template=False))
        except IOError:
            pass
        _temp_file.close()
        os.remove(_temp_file.name)
    logger.debug("matplotlib backend: %s", mpl.get_backend())
except AttributeError:  # mock module
    pass
